# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import ssl
import copy
import json
import time
import socket
import binascii
from typing import Any, Dict, Type, Union, Optional

import libcloud
from libcloud.http import LibcloudConnection, HttpLibResponseProxy
from libcloud.utils.py3 import ET, httplib, urlparse, urlencode
from libcloud.utils.misc import lowercase_keys
from libcloud.utils.retry import Retry
from libcloud.common.types import LibcloudError, MalformedResponseError
from libcloud.common.exceptions import exception_from_message

__all__ = [
    "RETRY_FAILED_HTTP_REQUESTS",
    "BaseDriver",
    "Connection",
    "PollingConnection",
    "ConnectionKey",
    "ConnectionUserAndKey",
    "CertificateConnection",
    "Response",
    "HTTPResponse",
    "JsonResponse",
    "XmlResponse",
    "RawResponse",
]

# Module level variable indicates if the failed HTTP requests should be retried
RETRY_FAILED_HTTP_REQUESTS = False

# Set to True to allow double slashes in the URL path. This way
# morph_action_hook() won't strip potentially double slashes in the URLs.
# This is to support scenarios such as this one -
# https://github.com/apache/libcloud/issues/1529.
# We default it to False for backward compatibility reasons.
ALLOW_PATH_DOUBLE_SLASHES = False


class LazyObject:
    """An object that doesn't get initialized until accessed."""

    @classmethod
    def _proxy(cls, *lazy_init_args, **lazy_init_kwargs):
        class Proxy(cls):
            _lazy_obj = None

            def __init__(self):
                # Must override the lazy_cls __init__
                pass

            def __getattribute__(self, attr):
                lazy_obj = object.__getattribute__(self, "_get_lazy_obj")()
                return getattr(lazy_obj, attr)

            def __setattr__(self, attr, value):
                lazy_obj = object.__getattribute__(self, "_get_lazy_obj")()
                setattr(lazy_obj, attr, value)

            def _get_lazy_obj(self):
                lazy_obj = object.__getattribute__(self, "_lazy_obj")
                if lazy_obj is None:
                    lazy_obj = cls(*lazy_init_args, **lazy_init_kwargs)
                    object.__setattr__(self, "_lazy_obj", lazy_obj)
                return lazy_obj

        return Proxy()

    @classmethod
    def lazy(cls, *lazy_init_args, **lazy_init_kwargs):
        """Create a lazily instantiated instance of the subclass, cls."""
        return cls._proxy(*lazy_init_args, **lazy_init_kwargs)


class HTTPResponse(httplib.HTTPResponse):
    # On python 2.6 some calls can hang because HEAD isn't quite properly
    # supported.
    # In particular this happens on S3 when calls are made to get_object to
    # objects that don't exist.
    # This applies the behaviour from 2.7, fixing the hangs.
    def read(self, amt=None):
        if self.fp is None:
            return ""

        if self._method == "HEAD":
            self.close()
            return ""

        return httplib.HTTPResponse.read(self, amt)


class Response:
    """
    A base Response class to derive from.
    """

    # Response status code
    status = httplib.OK  # type: int
    # Response headers
    headers = {}  # type: dict

    # Raw response body
    body = None

    # Parsed response body
    object = None

    error = None  # Reason returned by the server.
    connection = None  # Parent connection class
    parse_zero_length_body = False

    def __init__(self, response, connection):
        """
        :param response: HTTP response object. (optional)
        :type response: :class:`httplib.HTTPResponse`

        :param connection: Parent connection object.
        :type connection: :class:`.Connection`
        """
        self.connection = connection

        # http.client In Python 3 doesn't automatically lowercase the header
        # names
        self.headers = lowercase_keys(dict(response.headers))
        self.error = response.reason
        self.status = response.status_code
        self.request = response.request
        self.iter_content = response.iter_content

        self.body = (
            response.text.strip()
            if response.text is not None and hasattr(response.text, "strip")
            else ""
        )

        if not self.success():
            raise exception_from_message(
                code=self.status, message=self.parse_error(), headers=self.headers
            )

        self.object = self.parse_body()

    def parse_body(self):
        """
        Parse response body.

        Override in a provider's subclass.

        :return: Parsed body.
        :rtype: ``str``
        """
        return self.body if self.body is not None else ""

    def parse_error(self):
        """
        Parse the error messages.

        Override in a provider's subclass.

        :return: Parsed error.
        :rtype: ``str``
        """
        return self.body

    def success(self):
        """
        Determine if our request was successful.

        The meaning of this can be arbitrary; did we receive OK status? Did
        the node get created? Were we authenticated?

        :rtype: ``bool``
        :return: ``True`` or ``False``
        """
        # pylint: disable=E1101
        import requests

        return self.status in [
            requests.codes.ok,
            requests.codes.created,
            httplib.OK,
            httplib.CREATED,
            httplib.ACCEPTED,
        ]


class JsonResponse(Response):
    """
    A Base JSON Response class to derive from.
    """

    def parse_body(self):
        if len(self.body) == 0 and not self.parse_zero_length_body:
            return self.body

        try:
            body = json.loads(self.body)
        except Exception:
            raise MalformedResponseError(
                "Failed to parse JSON", body=self.body, driver=self.connection.driver
            )
        return body

    parse_error = parse_body


class XmlResponse(Response):
    """
    A Base XML Response class to derive from.
    """

    def parse_body(self):
        if len(self.body) == 0 and not self.parse_zero_length_body:
            return self.body

        try:
            try:
                body = ET.XML(self.body)
            except ValueError:
                # lxml wants a bytes and tests are basically hard-coded to str
                body = ET.XML(self.body.encode("utf-8"))
        except Exception:
            raise MalformedResponseError(
                "Failed to parse XML", body=self.body, driver=self.connection.driver
            )
        return body

    parse_error = parse_body


class RawResponse(Response):
    def __init__(self, connection, response=None):
        """
        :param connection: Parent connection object.
        :type connection: :class:`.Connection`
        """
        self._status = None
        self._response = None
        self._headers = {}
        self._error = None
        self._reason = None
        self.connection = connection
        if response is not None:
            self.headers = lowercase_keys(dict(response.headers))
            self.error = response.reason
            self.status = response.status_code
            self.request = response.request
            self.iter_content = response.iter_content

        if not self.success():
            self.parse_error()

    def success(self):
        """
        Determine if our request was successful.

        The meaning of this can be arbitrary; did we receive OK status? Did
        the node get created? Were we authenticated?

        :rtype: ``bool``
        :return: ``True`` or ``False``
        """
        # pylint: disable=E1101
        import requests

        return self.status in [
            requests.codes.ok,
            requests.codes.created,
            httplib.OK,
            httplib.CREATED,
            httplib.ACCEPTED,
        ]

    @property
    def response(self):
        if not self._response:
            response = self.connection.connection.getresponse()
            self._response = HttpLibResponseProxy(response)
            if not self.success():
                self.parse_error()
        return self._response

    @property
    def body(self):
        # Note: We use property to avoid saving whole response body into RAM
        # See https://github.com/apache/libcloud/pull/1132 for details
        return self.response.body

    @property
    def reason(self):
        if not self._reason:
            self._reason = self.response.reason
        return self._reason


class Connection:
    """
    A Base Connection class to derive from.
    """

    conn_class = LibcloudConnection

    responseCls = Response
    rawResponseCls = RawResponse
    retryCls = Retry
    connection = None
    host = "127.0.0.1"  # type: str
    port = 443
    timeout = None  # type: Optional[Union[int, float]]
    secure = 1
    driver = None  # type:  Type[BaseDriver]
    action = None
    cache_busting = False
    backoff = None
    retry_delay = None

    allow_insecure = True

    def __init__(
        self,
        secure=True,
        host=None,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        retry_delay=None,
        backoff=None,
    ):
        self.secure = secure and 1 or 0
        self.ua = []
        self.context = {}

        if not self.allow_insecure and not secure:
            # TODO: We should eventually switch to whitelist instead of
            # blacklist approach
            raise ValueError("Non https connections are not allowed (use " "secure=True)")

        self.request_path = ""

        if host:
            self.host = host

        if port is not None:
            self.port = port
        else:
            if self.secure == 1:
                self.port = 443
            else:
                self.port = 80

        if url:
            (
                self.host,
                self.port,
                self.secure,
                self.request_path,
            ) = self._tuple_from_url(url)

        self.timeout = timeout or self.timeout
        self.retry_delay = retry_delay
        self.backoff = backoff
        self.proxy_url = proxy_url

    def set_http_proxy(self, proxy_url):
        """
        Set a HTTP / HTTPS proxy which will be used with this connection.

        :param proxy_url: Proxy URL (e.g. http://<hostname>:<port> without
                          authentication and
                          <scheme>://<username>:<password>@<hostname>:<port>
                          for basic auth authentication information.
        :type proxy_url: ``str``
        """
        self.proxy_url = proxy_url

        # NOTE: Because of the way connection instantion works, we need to call
        # self.connection.set_http_proxy() here. Just setting "self.proxy_url"
        # won't work.
        self.connection.set_http_proxy(proxy_url=proxy_url)

    def set_context(self, context):
        if not isinstance(context, dict):
            raise TypeError("context needs to be a dictionary")

        self.context = context

    def reset_context(self):
        self.context = {}

    def _tuple_from_url(self, url):
        secure = 1
        port = None
        (scheme, netloc, request_path, param, query, fragment) = urlparse.urlparse(url)

        if scheme not in ["http", "https"]:
            raise LibcloudError("Invalid scheme: {} in url {}".format(scheme, url))

        if scheme == "http":
            secure = 0

        if ":" in netloc:
            netloc, port = netloc.rsplit(":")
            port = int(port)

        if not port:
            if scheme == "http":
                port = 80
            else:
                port = 443

        host = netloc
        port = int(port)

        return (host, port, secure, request_path)

    def connect(self, host=None, port=None, base_url=None, **kwargs):
        """
        Establish a connection with the API server.

        :type host: ``str``
        :param host: Optional host to override our default

        :type port: ``int``
        :param port: Optional port to override our default

        :returns: A connection
        """
        # prefer the attribute base_url if its set or sent
        connection = None
        secure = self.secure

        if getattr(self, "base_url", None) and base_url is None:
            (host, port, secure, request_path) = self._tuple_from_url(getattr(self, "base_url"))
        elif base_url is not None:
            (host, port, secure, request_path) = self._tuple_from_url(base_url)
        else:
            host = host or self.host
            port = port or self.port

        # Make sure port is an int
        port = int(port)

        if not hasattr(kwargs, "host"):
            kwargs.update({"host": host})

        if not hasattr(kwargs, "port"):
            kwargs.update({"port": port})

        if not hasattr(kwargs, "secure"):
            kwargs.update({"secure": self.secure})

        if not hasattr(kwargs, "key_file") and hasattr(self, "key_file"):
            kwargs.update({"key_file": getattr(self, "key_file")})

        if not hasattr(kwargs, "cert_file") and hasattr(self, "cert_file"):
            kwargs.update({"cert_file": getattr(self, "cert_file")})

        if self.timeout:
            kwargs.update({"timeout": self.timeout})

        if self.proxy_url:
            kwargs.update({"proxy_url": self.proxy_url})

        connection = self.conn_class(**kwargs)
        # You can uncomment this line, if you setup a reverse proxy server
        # which proxies to your endpoint, and lets you easily capture
        # connections in cleartext when you setup the proxy to do SSL
        # for you
        # connection = self.conn_class("127.0.0.1", 8080)

        self.connection = connection

    def _user_agent(self):
        user_agent_suffix = " ".join(["(%s)" % x for x in self.ua])

        if self.driver:
            user_agent = "libcloud/{} ({}) {}".format(
                libcloud.__version__,
                self.driver.name,
                user_agent_suffix,
            )
        else:
            user_agent = "libcloud/{} {}".format(libcloud.__version__, user_agent_suffix)

        return user_agent

    def user_agent_append(self, token):
        """
        Append a token to a user agent string.

        Users of the library should call this to uniquely identify their
        requests to a provider.

        :type token: ``str``
        :param token: Token to add to the user agent.
        """
        self.ua.append(token)

    def request(
        self,
        action,
        params=None,
        data=None,
        headers=None,
        method="GET",
        raw=False,
        stream=False,
        json=None,
        retry_failed=None,
    ):
        """
        Request a given `action`.

        Basically a wrapper around the connection
        object's `request` that does some helpful pre-processing.

        :type action: ``str``
        :param action: A path. This can include arguments. If included,
            any extra parameters are appended to the existing ones.

        :type params: ``dict``
        :param params: Optional mapping of additional parameters to send. If
            None, leave as an empty ``dict``.

        :type data: ``unicode``
        :param data: A body of data to send with the request.

        :type headers: ``dict``
        :param headers: Extra headers to add to the request
            None, leave as an empty ``dict``.

        :type method: ``str``
        :param method: An HTTP method such as "GET" or "POST".

        :type raw: ``bool``
        :param raw: True to perform a "raw" request aka only send the headers
                     and use the rawResponseCls class. This is used with
                     storage API when uploading a file.

        :type stream: ``bool``
        :param stream: True to return an iterator in Response.iter_content
                    and allow streaming of the response data
                    (for downloading large files)

        :param retry_failed: True if failed requests should be retried. This
                              argument can override module level constant and
                              environment variable value on per-request basis.

        :return: An :class:`Response` instance.
        :rtype: :class:`Response` instance

        """
        if params is None:
            params = {}
        else:
            params = copy.copy(params)

        if headers is None:
            headers = {}
        else:
            headers = copy.copy(headers)

        retry_enabled = (
            os.environ.get("LIBCLOUD_RETRY_FAILED_HTTP_REQUESTS", False)
            or RETRY_FAILED_HTTP_REQUESTS
        )

        # Method level argument has precedence over module level constant and
        # environment variable
        if retry_failed is not None:
            retry_enabled = retry_failed

        action = self.morph_action_hook(action)
        self.action = action
        self.method = method
        self.data = data

        # Extend default parameters
        params = self.add_default_params(params)

        # Add cache busting parameters (if enabled)
        if self.cache_busting and method == "GET":
            params = self._add_cache_busting_to_params(params=params)

        # Extend default headers
        headers = self.add_default_headers(headers)

        # We always send a user-agent header
        headers.update({"User-Agent": self._user_agent()})

        # Indicate that we support gzip and deflate compression
        headers.update({"Accept-Encoding": "gzip,deflate"})

        port = int(self.port)

        if port not in (80, 443):
            headers.update({"Host": "%s:%d" % (self.host, port)})
        else:
            headers.update({"Host": self.host})

        if data:
            data = self.encode_data(data)

        params, headers = self.pre_connect_hook(params, headers)

        if params:
            if "?" in action:
                url = "&".join((action, urlencode(params, doseq=True)))
            else:
                url = "?".join((action, urlencode(params, doseq=True)))
        else:
            url = action

        # IF connection has not yet been established
        if self.connection is None:
            self.connect()

        request_to_be_executed = self._retryable_request

        if retry_enabled:
            retry_request = self.retryCls(
                retry_delay=self.retry_delay, timeout=self.timeout, backoff=self.backoff
            )
            request_to_be_executed = retry_request(self._retryable_request)

        return request_to_be_executed(
            url=url, method=method, raw=raw, stream=stream, headers=headers, data=data
        )

    def _retryable_request(
        self,
        url: str,
        data: bytes,
        headers: Dict[str, Any],
        method: str,
        raw: bool,
        stream: bool,
    ) -> Union[RawResponse, Response]:
        try:
            # @TODO: Should we just pass File object as body to request method
            # instead of dealing with splitting and sending the file ourselves?
            assert self.connection is not None

            if raw:
                self.connection.prepared_request(
                    method=method,
                    url=url,
                    body=data,
                    headers=headers,
                    raw=raw,
                    stream=stream,
                )
            else:
                self.connection.request(
                    method=method, url=url, body=data, headers=headers, stream=stream
                )

        except socket.gaierror as e:
            message = str(e)
            errno = getattr(e, "errno", None)

            if errno == -5:
                # Throw a more-friendly exception on "no address associated
                # with hostname" error. This error could simply indicate that
                # "host" Connection class attribute is set to an incorrect
                # value
                class_name = self.__class__.__name__
                msg = (
                    '%s. Perhaps "host" Connection class attribute '
                    "(%s.connection) is set to an invalid, non-hostname "
                    "value (%s)?" % (message, class_name, self.host)
                )
                raise socket.gaierror(msg)  # type: ignore
            self.reset_context()
            raise e
        except ssl.SSLError as e:
            self.reset_context()
            raise ssl.SSLError(str(e))

        if raw:
            responseCls = self.rawResponseCls
            kwargs = {"connection": self, "response": self.connection.getresponse()}
        else:
            responseCls = self.responseCls
            kwargs = {"connection": self, "response": self.connection.getresponse()}

        try:
            response = responseCls(**kwargs)
        finally:
            # Always reset the context after the request has completed
            self.reset_context()

        return response

    def morph_action_hook(self, action):
        """
        Here we strip any duplicated leading or trailing slashes to
        prevent typos and other issues where some APIs don't correctly
        handle double slashes.

        Keep in mind that in some situations, "/" is a valid path name
        so we have a module flag which disables this behavior
        (https://github.com/apache/libcloud/issues/1529).
        """
        if ALLOW_PATH_DOUBLE_SLASHES:
            # Special case to support scenarios where double slashes are
            # valid - e.g. for S3 paths - /bucket//path1/path2.txt
            return self.request_path + action

        url = urlparse.urljoin(self.request_path.lstrip("/").rstrip("/") + "/", action.lstrip("/"))

        if not url.startswith("/"):
            return "/" + url
        else:
            return url

    def add_default_params(self, params):
        """
        Adds default parameters (such as API key, version, etc.)
        to the passed `params`

        Should return a dictionary.
        """
        return params

    def add_default_headers(self, headers):
        """
        Adds default headers (such as Authorization, X-Foo-Bar)
        to the passed `headers`

        Should return a dictionary.
        """
        return headers

    def pre_connect_hook(self, params, headers):
        """
        A hook which is called before connecting to the remote server.
        This hook can perform a final manipulation on the params, headers and
        url parameters.

        :type params: ``dict``
        :param params: Request parameters.

        :type headers: ``dict``
        :param headers: Request headers.
        """
        return params, headers

    def encode_data(self, data):
        """
        Encode body data.

        Override in a provider's subclass.
        """
        return data

    def _add_cache_busting_to_params(self, params):
        """
        Add cache busting parameter to the query parameters of a GET request.

        Parameters are only added if "cache_busting" class attribute is set to
        True.

        Note: This should only be used with *naughty* providers which use
        excessive caching of responses.
        """
        cache_busting_value = binascii.hexlify(os.urandom(8)).decode("ascii")

        if isinstance(params, dict):
            params["cache-busting"] = cache_busting_value
        else:
            params.append(("cache-busting", cache_busting_value))

        return params


class PollingConnection(Connection):
    """
    Connection class which can also work with the async APIs.

    After initial requests, this class periodically polls for jobs status and
    waits until the job has finished.
    If job doesn't finish in timeout seconds, an Exception thrown.
    """

    poll_interval = 0.5
    timeout = 200
    request_method = "request"

    def async_request(
        self, action, params=None, data=None, headers=None, method="GET", context=None
    ):
        """
        Perform an 'async' request to the specified path. Keep in mind that
        this function is *blocking* and 'async' in this case means that the
        hit URL only returns a job ID which is the periodically polled until
        the job has completed.

        This function works like this:

        - Perform a request to the specified path. Response should contain a
          'job_id'.

        - Returned 'job_id' is then used to construct a URL which is used for
          retrieving job status. Constructed URL is then periodically polled
          until the response indicates that the job has completed or the
          timeout of 'self.timeout' seconds has been reached.

        :type action: ``str``
        :param action: A path

        :type params: ``dict``
        :param params: Optional mapping of additional parameters to send. If
            None, leave as an empty ``dict``.

        :type data: ``unicode``
        :param data: A body of data to send with the request.

        :type headers: ``dict``
        :param headers: Extra headers to add to the request
            None, leave as an empty ``dict``.

        :type method: ``str``
        :param method: An HTTP method such as "GET" or "POST".

        :type context: ``dict``
        :param context: Context dictionary which is passed to the functions
                        which construct initial and poll URL.

        :return: An :class:`Response` instance.
        :rtype: :class:`Response` instance
        """

        request = getattr(self, self.request_method)
        kwargs = self.get_request_kwargs(
            action=action,
            params=params,
            data=data,
            headers=headers,
            method=method,
            context=context,
        )
        response = request(**kwargs)
        kwargs = self.get_poll_request_kwargs(
            response=response, context=context, request_kwargs=kwargs
        )

        end = time.time() + self.timeout
        completed = False
        while time.time() < end and not completed:
            response = request(**kwargs)
            completed = self.has_completed(response=response)
            if not completed:
                time.sleep(self.poll_interval)

        if not completed:
            raise LibcloudError("Job did not complete in %s seconds" % (self.timeout))

        return response

    def get_request_kwargs(
        self, action, params=None, data=None, headers=None, method="GET", context=None
    ):
        """
        Arguments which are passed to the initial request() call inside
        async_request.
        """
        kwargs = {
            "action": action,
            "params": params,
            "data": data,
            "headers": headers,
            "method": method,
        }
        return kwargs

    def get_poll_request_kwargs(self, response, context, request_kwargs):
        """
        Return keyword arguments which are passed to the request() method when
        polling for the job status.

        :param response: Response object returned by poll request.
        :type response: :class:`HTTPResponse`

        :param request_kwargs: Kwargs previously used to initiate the
                                  poll request.
        :type response: ``dict``

        :return ``dict`` Keyword arguments
        """
        raise NotImplementedError("get_poll_request_kwargs not implemented")

    def has_completed(self, response):
        """
        Return job completion status.

        :param response: Response object returned by poll request.
        :type response: :class:`HTTPResponse`

        :return ``bool`` True if the job has completed, False otherwise.
        """
        raise NotImplementedError("has_completed not implemented")


class ConnectionKey(Connection):
    """
    Base connection class which accepts a single ``key`` argument.
    """

    def __init__(
        self,
        key,
        secure=True,
        host=None,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        backoff=None,
        retry_delay=None,
    ):
        """
        Initialize `user_id` and `key`; set `secure` to an ``int`` based on
        passed value.
        """
        super().__init__(
            secure=secure,
            host=host,
            port=port,
            url=url,
            timeout=timeout,
            proxy_url=proxy_url,
            backoff=backoff,
            retry_delay=retry_delay,
        )
        self.key = key


class CertificateConnection(Connection):
    """
    Base connection class which accepts a single ``cert_file`` argument.
    """

    def __init__(
        self,
        cert_file,
        secure=True,
        host=None,
        port=None,
        url=None,
        proxy_url=None,
        timeout=None,
        backoff=None,
        retry_delay=None,
    ):
        """
        Initialize `cert_file`; set `secure` to an ``int`` based on
        passed value.
        """
        super().__init__(
            secure=secure,
            host=host,
            port=port,
            url=url,
            timeout=timeout,
            backoff=backoff,
            retry_delay=retry_delay,
            proxy_url=proxy_url,
        )

        self.cert_file = cert_file


class KeyCertificateConnection(CertificateConnection):
    """
    Base connection class which accepts both ``key_file`` and ``cert_file``
    argument.
    """

    key_file = None

    def __init__(
        self,
        key_file,
        cert_file,
        secure=True,
        host=None,
        port=None,
        url=None,
        proxy_url=None,
        timeout=None,
        backoff=None,
        retry_delay=None,
        ca_cert=None,
    ):
        """
        Initialize `cert_file`; set `secure` to an ``int`` based on
        passed value.
        """
        super().__init__(
            cert_file,
            secure=secure,
            host=host,
            port=port,
            url=url,
            timeout=timeout,
            backoff=backoff,
            retry_delay=retry_delay,
            proxy_url=proxy_url,
        )

        self.key_file = key_file


class ConnectionUserAndKey(ConnectionKey):
    """
    Base connection class which accepts a ``user_id`` and ``key`` argument.
    """

    user_id = None  # type: int

    def __init__(
        self,
        user_id,
        key,
        secure=True,
        host=None,
        port=None,
        url=None,
        timeout=None,
        proxy_url=None,
        backoff=None,
        retry_delay=None,
    ):
        super().__init__(
            key,
            secure=secure,
            host=host,
            port=port,
            url=url,
            timeout=timeout,
            backoff=backoff,
            retry_delay=retry_delay,
            proxy_url=proxy_url,
        )
        self.user_id = user_id


class BaseDriver:
    """
    Base driver class from which other classes can inherit from.
    """

    connectionCls = ConnectionKey  # type: Type[Connection]

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=None,
        region=None,
        **kwargs,
    ):
        """
        :param    key:    API key or username to be used (required)
        :type     key:    ``str``

        :param    secret: Secret password to be used (required)
        :type     secret: ``str``

        :param    secure: Whether to use HTTPS or HTTP. Note: Some providers
                            only support HTTPS, and it is on by default.
        :type     secure: ``bool``

        :param    host: Override hostname used for connections.
        :type     host: ``str``

        :param    port: Override port used for connections.
        :type     port: ``int``

        :param    api_version: Optional API version. Only used by drivers
                                 which support multiple API versions.
        :type     api_version: ``str``

        :param region: Optional driver region. Only used by drivers which
                       support multiple regions.
        :type region: ``str``

        :rtype: ``None``
        """

        self.key = key
        self.secret = secret
        self.secure = secure
        self.api_version = api_version
        self.region = region

        conn_kwargs = self._ex_connection_class_kwargs()
        conn_kwargs.update(
            {
                "timeout": kwargs.pop("timeout", None),
                "retry_delay": kwargs.pop("retry_delay", None),
                "backoff": kwargs.pop("backoff", None),
                "proxy_url": kwargs.pop("proxy_url", None),
            }
        )

        args = [self.key]

        if self.secret is not None:
            args.append(self.secret)

        args.append(secure)

        host = conn_kwargs.pop("host", None) or host

        if host is not None:
            args.append(host)

        if port is not None:
            args.append(port)

        self.connection = self.connectionCls(*args, **conn_kwargs)
        self.connection.driver = self
        self.connection.connect()

    def _ex_connection_class_kwargs(self):
        """
        Return extra connection keyword arguments which are passed to the
        Connection class constructor.
        """
        return {}



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hmac
import hashlib
from datetime import datetime

from libcloud.utils.py3 import urlquote

__all__ = [
    "OSCRequestSignerAlgorithmV4",
]


class OSCRequestSigner:
    """
    Class which handles signing the outgoing AWS requests.
    """

    def __init__(self, access_key: str, access_secret: str, version: str, connection):
        """
        :param access_key: Access key.
        :type access_key: ``str``

        :param access_secret: Access secret.
        :type access_secret: ``str``

        :param version: API version.
        :type version: ``str``

        :param connection: Connection instance.
        :type connection: :class:`Connection`
        """
        self.access_key = access_key
        self.access_secret = access_secret
        self.version = version
        self.connection = connection


class OSCRequestSignerAlgorithmV4(OSCRequestSigner):
    @staticmethod
    def sign(key, msg):
        return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()

    @staticmethod
    def _get_signed_headers(headers: dict):
        return ";".join([k.lower() for k in sorted(headers.keys())])

    @staticmethod
    def _get_canonical_headers(headers: dict):
        return (
            "\n".join([":".join([k.lower(), str(v).strip()]) for k, v in sorted(headers.items())])
            + "\n"
        )

    @staticmethod
    def _get_request_params(params: dict):
        return "&".join(
            [
                "{}={}".format(urlquote(k, safe=""), urlquote(str(v), safe="~"))
                for k, v in sorted(params.items())
            ]
        )

    def get_request_headers(self, service_name: str, region: str, action: str, data: str):
        date = datetime.utcnow()
        host = "{}.{}.outscale.com".format(service_name, region)
        headers = {
            "Content-Type": "application/json; charset=utf-8",
            "X-Osc-Date": date.strftime("%Y%m%dT%H%M%SZ"),
            "Host": host,
        }
        path = "/{}/{}/{}".format(self.connection.service_name, self.version, action)
        sig = self._get_authorization_v4_header(
            headers=headers, dt=date, method="POST", path=path, data=data
        )
        headers.update({"Authorization": sig})
        return headers

    def _get_authorization_v4_header(
        self,
        headers: dict,
        data: str,
        dt: datetime,
        method: str = "GET",
        path: str = "/",
    ):
        credentials_scope = self._get_credential_scope(dt=dt)
        signed_headers = self._get_signed_headers(headers=headers)
        signature = self._get_signature(headers=headers, dt=dt, method=method, path=path, data=data)
        return (
            "OSC4-HMAC-SHA256 Credential=%(u)s/%(c)s, "
            "SignedHeaders=%(sh)s, Signature=%(s)s"
            % {
                "u": self.access_key,
                "c": credentials_scope,
                "sh": signed_headers,
                "s": signature,
            }
        )

    def _get_signature(self, headers: dict, dt: datetime, method: str, path: str, data: str):
        string_to_sign = self._get_string_to_sign(
            headers=headers, dt=dt, method=method, path=path, data=data
        )
        signing_key = self._get_key_to_sign_with(self.access_secret, dt.strftime("%Y%m%d"))
        return hmac.new(signing_key, string_to_sign.encode("utf-8"), hashlib.sha256).hexdigest()

    def _get_key_to_sign_with(self, key: str, dt: str):
        k_date = self.sign(("OSC4" + key).encode("utf-8"), dt)
        k_region = self.sign(k_date, self.connection.region_name)
        k_service = self.sign(k_region, self.connection.service_name)
        return self.sign(k_service, "osc4_request")

    def _get_string_to_sign(self, headers: dict, dt: datetime, method: str, path: str, data: str):
        canonical_request = self._get_canonical_request(
            headers=headers, method=method, path=path, data=data
        )
        return (
            "OSC4-HMAC-SHA256"
            + "\n"
            + dt.strftime("%Y%m%dT%H%M%SZ")
            + "\n"
            + self._get_credential_scope(dt)
            + "\n"
            + hashlib.sha256(canonical_request.encode("utf-8")).hexdigest()
        )

    def _get_credential_scope(self, dt):
        return "/".join(
            [
                dt.strftime("%Y%m%d"),
                self.connection.region_name,
                self.connection.service_name,
                "osc4_request",
            ]
        )

    def _get_canonical_request(self, headers, method, path, data):
        return "\n".join(
            [
                method,
                path,
                self._get_request_params({}),
                self._get_canonical_headers(headers),
                self._get_signed_headers(headers),
                hashlib.sha256(data.encode("utf-8")).hexdigest(),
            ]
        )



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
World Wide DNS Driver
"""

__all__ = ["WorldWideDNSDriver"]

import re

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import (
    Provider,
    RecordType,
    RecordError,
    ZoneDoesNotExistError,
    RecordDoesNotExistError,
)
from libcloud.common.types import LibcloudError
from libcloud.common.worldwidedns import WorldWideDNSConnection

MAX_RECORD_ENTRIES = 40  # Maximum record entries for zone


class WorldWideDNSError(LibcloudError):
    def __repr__(self):
        return "<WorldWideDNSError in " + repr(self.driver) + " " + repr(self.value) + ">"


class WorldWideDNSDriver(DNSDriver):
    type = Provider.WORLDWIDEDNS
    name = "World Wide DNS"
    website = "https://www.worldwidedns.net/"
    connectionCls = WorldWideDNSConnection

    RECORD_TYPE_MAP = {
        RecordType.MX: "MX",
        RecordType.CNAME: "CNAME",
        RecordType.A: "A",
        RecordType.NS: "NS",
        RecordType.SRV: "SRV",
        RecordType.TXT: "TXT",
    }

    def __init__(
        self,
        key,
        secret=None,
        reseller_id=None,
        secure=True,
        host=None,
        port=None,
        **kwargs,
    ):
        """
        :param    key: API key or username to used (required)
        :type     key: ``str``

        :param    secret: Secret password to be used (required)
        :type     secret: ``str``

        :param    reseller_id: Reseller ID for reseller accounts
        :type     reseller_id: ``str``

        :param    secure: Whether to use HTTPS or HTTP. Note: Some providers
                          only support HTTPS, and it is on by default.
        :type     secure: ``bool``

        :param    host: Override hostname used for connections.
        :type     host: ``str``

        :param    port: Override port used for connections.
        :type     port: ``int``

        :return: ``None``
        """
        super().__init__(key=key, secret=secret, secure=secure, host=host, port=port, **kwargs)
        self.reseller_id = reseller_id

    def list_zones(self):
        """
        Return a list of zones.

        :return: ``list`` of :class:`Zone`

        For more info, please see:
        https://www.worldwidedns.net/dns_api_protocol_list.asp
        or
        https://www.worldwidedns.net/dns_api_protocol_list_reseller.asp
        """
        action = "/api_dns_list.asp"
        if self.reseller_id is not None:
            action = "/api_dns_list_reseller.asp"
        zones = self.connection.request(action)
        if len(zones.body) == 0:
            return []
        else:
            return self._to_zones(zones.body)

    def iterate_records(self, zone):
        """
        Return a generator to iterate over records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :rtype: ``generator`` of :class:`Record`
        """
        records = self._to_records(zone)
        yield from records

    def get_zone(self, zone_id):
        """
        Return a Zone instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :rtype: :class:`Zone`
        """
        zones = self.list_zones()
        zone = [zone for zone in zones if zone.id == zone_id]
        if len(zone) == 0:
            raise ZoneDoesNotExistError(
                driver=self, value="The zone doesn't exists", zone_id=zone_id
            )
        return zone[0]

    def get_record(self, zone_id, record_id):
        """
        Return a Record instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :param record_id: ID number of the required record.
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        zone = self.get_zone(zone_id)
        try:
            if int(record_id) not in range(1, MAX_RECORD_ENTRIES + 1):
                raise RecordDoesNotExistError(
                    value="Record doesn't exists",
                    driver=zone.driver,
                    record_id=record_id,
                )
        except ValueError:
            raise WorldWideDNSError(value="Record id should be a string number", driver=self)
        subdomain = zone.extra.get("S%s" % record_id)
        type = zone.extra.get("T%s" % record_id)
        data = zone.extra.get("D%s" % record_id)
        record = self._to_record(record_id, subdomain, type, data, zone)
        return record

    def update_zone(self, zone, domain, type="master", ttl=None, extra=None, ex_raw=False):
        """
        Update an existing zone.

        :param zone: Zone to update.
        :type  zone: :class:`Zone`

        :param domain: Zone domain name (e.g. example.com)
        :type  domain: ``str``

        :param type: Zone type (master / slave).
        :type  type: ``str``

        :param ttl: TTL for new records. (optional)
        :type  ttl: ``int``

        :param extra: Extra attributes (driver specific) (optional). Values not
                      specified such as *SECURE*, *IP*, *FOLDER*, *HOSTMASTER*,
                      *REFRESH*, *RETRY* and *EXPIRE* will be kept as already
                      is. The same will be for *S(1 to 40)*, *T(1 to 40)* and
                      *D(1 to 40)* if not in raw mode and for *ZONENS* and
                      *ZONEDATA* if it is.
        :type  extra: ``dict``

        :param ex_raw: Mode we use to do the update using zone file or not.
        :type  ex_raw: ``bool``

        :rtype: :class:`Zone`

        For more info, please see
        https://www.worldwidedns.net/dns_api_protocol_list_domain.asp
        or
        https://www.worldwidedns.net/dns_api_protocol_list_domain_raw.asp
        or
        https://www.worldwidedns.net/dns_api_protocol_list_domain_reseller.asp
        or
        https://www.worldwidedns.net/dns_api_protocol_list_domain_raw_reseller.asp
        """
        if extra is not None:
            not_specified = [key for key in zone.extra.keys() if key not in extra.keys()]
        else:
            not_specified = zone.extra.keys()

        if ttl is None:
            ttl = zone.ttl

        params = {"DOMAIN": domain, "TTL": ttl}

        for key in not_specified:
            params[key] = zone.extra[key]
        if extra is not None:
            params.update(extra)
        if ex_raw:
            action = "/api_dns_modify_raw.asp"
            if self.reseller_id is not None:
                action = "/api_dns_modify_raw_reseller.asp"
            method = "POST"
        else:
            action = "/api_dns_modify.asp"
            if self.reseller_id is not None:
                action = "/api_dns_modify_reseller.asp"
            method = "GET"
        response = self.connection.request(action, params=params, method=method)  # noqa
        zone = self.get_zone(zone.id)
        return zone

    def update_record(self, record, name, type, data, extra=None):
        """
        Update an existing record.

        :param record: Record to update.
        :type  record: :class:`Record`

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param type: DNS record type (MX, CNAME, A, NS, SRV, TXT).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: Contains 'entry' Entry position (1 thru 40)
        :type extra: ``dict``

        :rtype: :class:`Record`
        """
        if (extra is None) or ("entry" not in extra):
            raise WorldWideDNSError(value="You must enter 'entry' parameter", driver=self)
        record_id = extra.get("entry")
        if name == "":
            name = "@"
        if type not in self.RECORD_TYPE_MAP:
            raise RecordError(
                value="Record type is not allowed",
                driver=record.zone.driver,
                record_id=name,
            )
        zone = record.zone
        extra = {
            "S%s" % record_id: name,
            "T%s" % record_id: type,
            "D%s" % record_id: data,
        }
        zone = self.update_zone(zone, zone.domain, extra=extra)
        record = self.get_record(zone.id, record_id)
        return record

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        Create a new zone.

        :param domain: Zone domain name (e.g. example.com)
        :type domain: ``str``

        :param type: Zone type (master / slave).
        :type  type: ``str``

        :param ttl: TTL for new records. (optional)
        :type  ttl: ``int``

        :param extra: Extra attributes (driver specific). (optional). Possible
                      parameter in here should be *DYN* which values should be
                      1 for standard and 2 for dynamic. Default is 1.
        :type extra: ``dict``

        :rtype: :class:`Zone`

        For more info, please see
        https://www.worldwidedns.net/dns_api_protocol_new_domain.asp
        or
        https://www.worldwidedns.net/dns_api_protocol_new_domain_reseller.asp
        """
        if type == "master":
            _type = 0
        elif type == "slave":
            _type = 1
        if extra:
            dyn = extra.get("DYN") or 1
        else:
            dyn = 1
        params = {"DOMAIN": domain, "TYPE": _type}
        action = "/api_dns_new_domain.asp"
        if self.reseller_id is not None:
            params["DYN"] = dyn
            action = "/api_dns_new_domain_reseller.asp"
        self.connection.request(action, params=params)
        zone = self.get_zone(domain)
        if ttl is not None:
            zone = self.update_zone(zone, zone.domain, ttl=ttl)
        return zone

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a new record.

        We can create 40 record per domain. If all slots are full, we can
        replace one of them by choosing a specific entry in ``extra`` argument.

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param zone: Zone where the requested record is created.
        :type  zone: :class:`Zone`

        :param type: DNS record type (MX, CNAME, A, NS, SRV, TXT).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: Contains 'entry' Entry position (1 thru 40)
        :type extra: ``dict``

        :rtype: :class:`Record`
        """
        if (extra is None) or ("entry" not in extra):
            # If no entry is specified, we look for an available one. If all
            # are full, raise error.
            record_id = self._get_available_record_entry(zone)
            if not record_id:
                raise WorldWideDNSError(value="All record entries are full", driver=zone.driver)
        else:
            record_id = extra.get("entry")
        if name == "":
            name = "@"
        if type not in self.RECORD_TYPE_MAP:
            raise RecordError(
                value="Record type is not allowed",
                driver=zone.driver,
                record_id=record_id,
            )
        extra = {
            "S%s" % record_id: name,
            "T%s" % record_id: type,
            "D%s" % record_id: data,
        }
        zone = self.update_zone(zone, zone.domain, extra=extra)
        record = self.get_record(zone.id, record_id)
        return record

    def delete_zone(self, zone):
        """
        Delete a zone.

        Note: This will delete all the records belonging to this zone.

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``

        For more information, please see
        https://www.worldwidedns.net/dns_api_protocol_delete_domain.asp
        or
        https://www.worldwidedns.net/dns_api_protocol_delete_domain_reseller.asp
        """
        params = {"DOMAIN": zone.domain}
        action = "/api_dns_delete_domain.asp"
        if self.reseller_id is not None:
            action = "/api_dns_delete_domain_reseller.asp"
        response = self.connection.request(action, params=params)
        return response.success()

    def delete_record(self, record):
        """
        Delete a record.

        :param record: Record to delete.
        :type  record: :class:`Record`

        :rtype: ``bool``
        """
        zone = record.zone
        for index in range(MAX_RECORD_ENTRIES):
            if record.name == zone.extra["S%s" % (index + 1)]:
                entry = index + 1
                break
        extra = {"S%s" % entry: "", "T%s" % entry: "NONE", "D%s" % entry: ""}
        self.update_zone(zone, zone.domain, extra=extra)
        return True

    def ex_view_zone(self, domain, name_server):
        """
        View zone file from a name server

        :param domain: Domain name.
        :type  domain: ``str``

        :param name_server: Name server to check. (1, 2 or 3)
        :type  name_server: ``int``

        :rtype: ``str``

        For more info, please see:
        https://www.worldwidedns.net/dns_api_protocol_viewzone.asp
        or
        https://www.worldwidedns.net/dns_api_protocol_viewzone_reseller.asp
        """
        params = {"DOMAIN": domain, "NS": name_server}
        action = "/api_dns_viewzone.asp"
        if self.reseller_id is not None:
            action = "/api_dns_viewzone_reseller.asp"
        response = self.connection.request(action, params=params)
        return response.object

    def ex_transfer_domain(self, domain, user_id):
        """
        This command will allow you, if you are a reseller, to change the
        userid on a domain name to another userid in your account ONLY if that
        new userid is already created.

        :param domain: Domain name.
        :type  domain: ``str``

        :param user_id: The new userid to connect to the domain name.
        :type  user_id: ``str``

        :rtype: ``bool``

        For more info, please see:
        https://www.worldwidedns.net/dns_api_protocol_transfer.asp
        """
        if self.reseller_id is None:
            raise WorldWideDNSError("This is not a reseller account", driver=self)
        params = {"DOMAIN": domain, "NEW_ID": user_id}
        response = self.connection.request("/api_dns_transfer.asp", params=params)
        return response.success()

    def _get_available_record_entry(self, zone):
        """Return an available entry to store a record."""
        entries = zone.extra
        for entry in range(1, MAX_RECORD_ENTRIES + 1):
            subdomain = entries.get("S%s" % entry)
            _type = entries.get("T%s" % entry)
            data = entries.get("D%s" % entry)
            if not any([subdomain, _type, data]):
                return entry
        return None

    def _to_zones(self, data):
        domain_list = re.split("\r?\n", data)
        zones = []
        for line in domain_list:
            zone = self._to_zone(line)
            zones.append(zone)

        return zones

    def _to_zone(self, line):
        data = line.split("\x1f")
        name = data[0]
        if data[1] == "P":
            type = "master"
            domain_data = self._get_domain_data(name)
            resp_lines = re.split("\r?\n", domain_data.body)
            soa_block = resp_lines[:6]
            zone_data = resp_lines[6:]
            extra = {
                "HOSTMASTER": soa_block[0],
                "REFRESH": soa_block[1],
                "RETRY": soa_block[2],
                "EXPIRE": soa_block[3],
                "SECURE": soa_block[5],
            }
            ttl = soa_block[4]
            for line in range(MAX_RECORD_ENTRIES):
                line_data = zone_data[line].split("\x1f")
                extra["S%s" % (line + 1)] = line_data[0]
                _type = line_data[1]
                extra["T%s" % (line + 1)] = _type if _type != "NONE" else ""
                try:
                    extra["D%s" % (line + 1)] = line_data[2]
                except IndexError:
                    extra["D%s" % (line + 1)] = ""
        elif data[1] == "S":
            type = "slave"
            extra = {}
            ttl = 0
        return Zone(id=name, domain=name, type=type, ttl=ttl, driver=self, extra=extra)

    def _get_domain_data(self, name):
        params = {"DOMAIN": name}
        data = self.connection.request("/api_dns_list_domain.asp", params=params)
        return data

    def _to_records(self, zone):
        records = []
        for record_id in range(1, MAX_RECORD_ENTRIES + 1):
            subdomain = zone.extra["S%s" % (record_id)]
            type = zone.extra["T%s" % (record_id)]
            data = zone.extra["D%s" % (record_id)]
            if subdomain and type and data:
                record = self._to_record(record_id, subdomain, type, data, zone)
                records.append(record)
        return records

    def _to_record(self, _id, subdomain, type, data, zone):
        return Record(id=_id, name=subdomain, type=type, data=data, zone=zone, driver=zone.driver)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common settings and connection objects for DigitalOcean Cloud
"""

from libcloud.utils.py3 import httplib, parse_qs, urlparse
from libcloud.common.base import BaseDriver, JsonResponse, ConnectionKey
from libcloud.common.types import LibcloudError, InvalidCredsError

__all__ = [
    "DigitalOcean_v2_Response",
    "DigitalOcean_v2_Connection",
    "DigitalOceanBaseDriver",
]


class DigitalOcean_v1_Error(LibcloudError):
    """
    Exception for when attempting to use version 1
    of the DigitalOcean API which is no longer
    supported.
    """

    def __init__(
        self,
        value=(
            "Driver no longer supported: Version 1 of the "
            "DigitalOcean API reached end of life on November 9, "
            "2015. Use the v2 driver. Please visit: "
            "https://developers.digitalocean.com/documentation/changelog/api-v1/sunsetting-api-v1/"
        ),  # noqa: E501
        driver=None,
    ):
        super().__init__(value, driver=driver)


class DigitalOcean_v2_Response(JsonResponse):
    valid_response_codes = [
        httplib.OK,
        httplib.ACCEPTED,
        httplib.CREATED,
        httplib.NO_CONTENT,
    ]

    def parse_error(self):
        if self.status == httplib.UNAUTHORIZED:
            body = self.parse_body()
            raise InvalidCredsError(body["message"])
        else:
            body = self.parse_body()
            if "message" in body:
                error = "{} (code: {})".format(body["message"], self.status)
            else:
                error = body
            return error

    def success(self):
        return self.status in self.valid_response_codes


class DigitalOcean_v2_Connection(ConnectionKey):
    """
    Connection class for the DigitalOcean (v2) driver.
    """

    host = "api.digitalocean.com"
    responseCls = DigitalOcean_v2_Response

    def add_default_headers(self, headers):
        """
        Add headers that are necessary for every request

        This method adds ``token`` to the request.
        """
        headers["Authorization"] = "Bearer %s" % (self.key)
        headers["Content-Type"] = "application/json"
        return headers

    def add_default_params(self, params):
        """
        Add parameters that are necessary for every request

        This method adds ``per_page`` to the request to reduce the total
        number of paginated requests to the API.
        """
        # pylint: disable=maybe-no-member
        params["per_page"] = self.driver.ex_per_page
        return params


class DigitalOceanConnection(DigitalOcean_v2_Connection):
    """
    Connection class for the DigitalOcean driver.
    """

    pass


class DigitalOceanResponse(DigitalOcean_v2_Response):
    pass


class DigitalOceanBaseDriver(BaseDriver):
    """
    DigitalOcean BaseDriver
    """

    name = "DigitalOcean"
    website = "https://www.digitalocean.com"

    def __new__(cls, key, secret=None, api_version="v2", **kwargs):
        if cls is DigitalOceanBaseDriver:
            if api_version == "v1" or secret is not None:
                raise DigitalOcean_v1_Error()
            elif api_version == "v2":
                cls = DigitalOcean_v2_BaseDriver
            else:
                raise NotImplementedError("Unsupported API version: %s" % (api_version))
        return super().__new__(cls, **kwargs)

    def ex_account_info(self):
        raise NotImplementedError("ex_account_info not implemented for this driver")

    def ex_list_events(self):
        raise NotImplementedError("ex_list_events not implemented for this driver")

    def ex_get_event(self, event_id):
        raise NotImplementedError("ex_get_event not implemented for this driver")

    def _paginated_request(self, url, obj):
        raise NotImplementedError("_paginated_requests not implemented for this driver")


class DigitalOcean_v2_BaseDriver(DigitalOceanBaseDriver):
    """
    DigitalOcean BaseDriver using v2 of the API.

    Supports `ex_per_page` ``int`` value keyword parameter to adjust per page
    requests against the API.
    """

    connectionCls = DigitalOcean_v2_Connection

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=None,
        region=None,
        ex_per_page=200,
        **kwargs,
    ):
        self.ex_per_page = ex_per_page
        super().__init__(key, **kwargs)

    def ex_account_info(self):
        return self.connection.request("/v2/account").object["account"]

    def ex_list_events(self):
        return self._paginated_request("/v2/actions", "actions")

    def ex_get_event(self, event_id):
        """
        Get an event object

        :param      event_id: Event id (required)
        :type       event_id: ``str``
        """
        params = {}
        return self.connection.request("/v2/actions/%s" % event_id, params=params).object["action"]

    def _paginated_request(self, url, obj):
        """
        Perform multiple calls in order to have a full list of elements when
        the API responses are paginated.

        :param url: API endpoint
        :type url: ``str``

        :param obj: Result object key
        :type obj: ``str``

        :return: ``list`` of API response objects
        :rtype: ``list``
        """
        params = {}
        data = self.connection.request(url)
        try:
            query = urlparse.urlparse(data.object["links"]["pages"]["last"])
            # The query[4] references the query parameters from the url
            pages = parse_qs(query[4])["page"][0]
            values = data.object[obj]
            for page in range(2, int(pages) + 1):
                params.update({"page": page})
                new_data = self.connection.request(url, params=params)

                more_values = new_data.object[obj]
                for value in more_values:
                    values.append(value)
            data = values
        except KeyError:  # No pages.
            data = data.object[obj]
        return data



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcloud.common.base import BaseDriver, ConnectionKey
from libcloud.common.types import LibcloudError

__all__ = ["Member", "LoadBalancer", "Algorithm", "Driver", "DEFAULT_ALGORITHM"]


class Member:
    """
    Represents a load balancer member.
    """

    def __init__(self, id, ip, port, balancer=None, extra=None):
        """
        :param id: Member ID.
        :type id: ``str``

        :param ip: IP address of this member.
        :param ip: ``str``

        :param port: Port of this member
        :param port: ``str``

        :param balancer: Balancer this member is attached to. (optional)
        :param balancer: :class:`.LoadBalancer`

        :param extra: Provider specific attributes.
        :type extra: ``dict``
        """
        self.id = str(id) if id else None
        self.ip = ip
        self.port = port
        self.balancer = balancer
        self.extra = extra or {}

    def __repr__(self):
        return "<Member: id={}, address={}:{}>".format(self.id, self.ip, self.port)


class LoadBalancer:
    """
    Provide a common interface for handling Load Balancers.
    """

    def __init__(self, id, name, state, ip, port, driver, extra=None):
        """
        :param id: Load balancer ID.
        :type id: ``str``

        :param name: Load balancer name.
        :type name: ``str``

        :param state: State this loadbalancer is in.
        :type state: :class:`libcloud.loadbalancer.types.State`

        :param ip: IP address of this loadbalancer.
        :type ip: ``str``

        :param port: Port of this loadbalancer.
        :type port: ``int``

        :param driver: Driver this loadbalancer belongs to.
        :type driver: :class:`.Driver`

        :param extra: Provider specific attributes. (optional)
        :type extra: ``dict``
        """
        self.id = str(id) if id else None
        self.name = name
        self.state = state
        self.ip = ip
        self.port = port
        self.driver = driver
        self.extra = extra or {}

    def attach_compute_node(self, node):
        return self.driver.balancer_attach_compute_node(balancer=self, node=node)

    def attach_member(self, member):
        return self.driver.balancer_attach_member(balancer=self, member=member)

    def detach_member(self, member):
        return self.driver.balancer_detach_member(balancer=self, member=member)

    def list_members(self):
        return self.driver.balancer_list_members(balancer=self)

    def destroy(self):
        return self.driver.destroy_balancer(balancer=self)

    def __repr__(self):
        return "<LoadBalancer: id=%s, name=%s, state=%s, ip=%s, " "port=%s>" % (
            self.id,
            self.name,
            self.state,
            self.ip,
            self.port,
        )


class Algorithm:
    """
    Represents a load balancing algorithm.
    """

    RANDOM = 0
    ROUND_ROBIN = 1
    LEAST_CONNECTIONS = 2
    WEIGHTED_ROUND_ROBIN = 3
    WEIGHTED_LEAST_CONNECTIONS = 4
    SHORTEST_RESPONSE = 5
    PERSISTENT_IP = 6
    LEAST_CONNECTIONS_MEMBER = 7
    LEAST_CONNECTIONS_NODE = 8
    OBSERVED_MEMBER = 9
    OBSERVED_NODE = 10
    PREDICTIVE_MEMBER = 11
    PREDICTIVE_NODE = 12


DEFAULT_ALGORITHM = Algorithm.ROUND_ROBIN


class Driver(BaseDriver):
    """
    A base Driver class to derive from

    This class is always subclassed by a specific driver.
    """

    name = None
    website = None

    connectionCls = ConnectionKey
    _ALGORITHM_TO_VALUE_MAP = {}
    _VALUE_TO_ALGORITHM_MAP = {}

    def __init__(self, key, secret=None, secure=True, host=None, port=None, **kwargs):
        super().__init__(key=key, secret=secret, secure=secure, host=host, port=port, **kwargs)

    def list_protocols(self):
        """
        Return a list of supported protocols.

        :rtype: ``list`` of ``str``
        """
        raise NotImplementedError("list_protocols not implemented for this driver")

    def list_balancers(self):
        """
        List all loadbalancers

        :rtype: ``list`` of :class:`LoadBalancer`
        """
        raise NotImplementedError("list_balancers not implemented for this driver")

    def create_balancer(self, name, port, protocol, algorithm, members):
        """
        Create a new load balancer instance

        :param name: Name of the new load balancer (required)
        :type  name: ``str``

        :param port: Port the load balancer should listen on, defaults to 80
        :type  port: ``str``

        :param protocol: Loadbalancer protocol, defaults to http.
        :type  protocol: ``str``

        :param members: list of Members to attach to balancer
        :type  members: ``list`` of :class:`Member`

        :param algorithm: Load balancing algorithm, defaults to ROUND_ROBIN.
        :type algorithm: :class:`.Algorithm`

        :rtype: :class:`LoadBalancer`
        """
        raise NotImplementedError("create_balancer not implemented for this driver")

    def destroy_balancer(self, balancer):
        """
        Destroy a load balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :return: ``True`` if the destroy was successful, otherwise ``False``.
        :rtype: ``bool``
        """

        raise NotImplementedError("destroy_balancer not implemented for this driver")

    def get_balancer(self, balancer_id):
        """
        Return a :class:`LoadBalancer` object.

        :param balancer_id: id of a load balancer you want to fetch
        :type  balancer_id: ``str``

        :rtype: :class:`LoadBalancer`
        """

        raise NotImplementedError("get_balancer not implemented for this driver")

    def update_balancer(self, balancer, **kwargs):
        """
        Sets the name, algorithm, protocol, or port on a load balancer.

        :param   balancer: LoadBalancer which should be used
        :type    balancer: :class:`LoadBalancer`

        :param name: New load balancer name
        :type    name: ``str``

        :param algorithm: New load balancer algorithm
        :type    algorithm: :class:`.Algorithm`

        :param protocol: New load balancer protocol
        :type    protocol: ``str``

        :param port: New load balancer port
        :type    port: ``int``

        :rtype: :class:`LoadBalancer`
        """
        raise NotImplementedError("update_balancer not implemented for this driver")

    def balancer_attach_compute_node(self, balancer, node):
        """
        Attach a compute node as a member to the load balancer.

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param node: Node to join to the balancer
        :type  node: :class:`Node`

        :return: Member after joining the balancer.
        :rtype: :class:`Member`
        """

        member = Member(id=None, ip=node.public_ips[0], port=balancer.port)
        return self.balancer_attach_member(balancer, member)

    def balancer_attach_member(self, balancer, member):
        """
        Attach a member to balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param member: Member to join to the balancer
        :type member: :class:`Member`

        :return: Member after joining the balancer.
        :rtype: :class:`Member`
        """

        raise NotImplementedError("balancer_attach_member not implemented for this driver")

    def balancer_detach_member(self, balancer, member):
        """
        Detach member from balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :param member: Member which should be used
        :type member: :class:`Member`

        :return: ``True`` if member detach was successful, otherwise ``False``.
        :rtype: ``bool``
        """

        raise NotImplementedError("balancer_detach_member not implemented for this driver")

    def balancer_list_members(self, balancer):
        """
        Return list of members attached to balancer

        :param balancer: LoadBalancer which should be used
        :type  balancer: :class:`LoadBalancer`

        :rtype: ``list`` of :class:`Member`
        """

        raise NotImplementedError("balancer_list_members not implemented for this driver")

    def list_supported_algorithms(self):
        """
        Return algorithms supported by this driver.

        :rtype: ``list`` of ``str``
        """
        return list(self._ALGORITHM_TO_VALUE_MAP.keys())

    def _value_to_algorithm(self, value):
        """
        Return :class:`.Algorithm` based on the value.

        :param value: Algorithm name (e.g. http, tcp, ...).
        :type  value: ``str``

        :rtype: :class:`.Algorithm`
        """
        try:
            return self._VALUE_TO_ALGORITHM_MAP[value]
        except KeyError:
            raise LibcloudError(value="Invalid value: %s" % (value), driver=self)

    def _algorithm_to_value(self, algorithm):
        """
        Return string value for the provided algorithm.

        :param value: Algorithm enum.
        :type  value: :class:`Algorithm`

        :rtype: ``str``
        """
        try:
            return self._ALGORITHM_TO_VALUE_MAP[algorithm]
        except KeyError:
            raise LibcloudError(value="Invalid algorithm: %s" % (algorithm), driver=self)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module which contains common Kubernetes related code.
"""

import os
import base64
import warnings
from typing import Optional

from libcloud.utils.py3 import b, httplib
from libcloud.common.base import (
    JsonResponse,
    ConnectionKey,
    ConnectionUserAndKey,
    KeyCertificateConnection,
)
from libcloud.common.types import InvalidCredsError

__all__ = [
    "KubernetesException",
    "KubernetesBasicAuthConnection",
    "KubernetesTLSAuthConnection",
    "KubernetesTokenAuthConnection",
    "KubernetesDriverMixin",
    "VALID_RESPONSE_CODES",
]

VALID_RESPONSE_CODES = [
    httplib.OK,
    httplib.ACCEPTED,
    httplib.CREATED,
    httplib.NO_CONTENT,
]


class KubernetesException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        self.args = (code, message)

    def __str__(self):
        return "{} {}".format(self.code, self.message)

    def __repr__(self):
        return "KubernetesException {} {}".format(self.code, self.message)


class KubernetesResponse(JsonResponse):
    valid_response_codes = [
        httplib.OK,
        httplib.ACCEPTED,
        httplib.CREATED,
        httplib.NO_CONTENT,
    ]

    def parse_error(self):
        if self.status == 401:
            raise InvalidCredsError("Invalid credentials")
        return self.body

    def success(self):
        return self.status in self.valid_response_codes


class KubernetesTLSAuthConnection(KeyCertificateConnection):
    responseCls = KubernetesResponse
    timeout = 60

    def __init__(
        self,
        key,
        secure=True,
        host="localhost",
        port="6443",
        key_file=None,
        cert_file=None,
        **kwargs,
    ):
        super().__init__(
            key_file=key_file,
            cert_file=cert_file,
            secure=secure,
            host=host,
            port=port,
            url=None,
            proxy_url=None,
            timeout=None,
            backoff=None,
            retry_delay=None,
        )

        if key_file:
            keypath = os.path.expanduser(key_file)
            is_file_path = os.path.exists(keypath) and os.path.isfile(keypath)

            if not is_file_path:
                raise InvalidCredsError(
                    "You need an key PEM file to authenticate "
                    "via tls. For more info please visit:"
                    "https://kubernetes.io/docs/concepts/"
                    "cluster-administration/certificates/"
                )

            self.key_file = key_file

            certpath = os.path.expanduser(cert_file)
            is_file_path = os.path.exists(certpath) and os.path.isfile(certpath)

            if not is_file_path:
                raise InvalidCredsError(
                    "You need an certificate PEM file to authenticate "
                    "via tls. For more info please visit:"
                    "https://kubernetes.io/docs/concepts/"
                    "cluster-administration/certificates/"
                )

            self.cert_file = cert_file

    def add_default_headers(self, headers):
        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"
        return headers


class KubernetesTokenAuthConnection(ConnectionKey):
    responseCls = KubernetesResponse
    timeout = 60

    def add_default_headers(self, headers):
        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"
        if self.key:
            headers["Authorization"] = "Bearer " + self.key
        else:
            raise ValueError("Please provide a valid token in the key param")
        return headers


class KubernetesBasicAuthConnection(ConnectionUserAndKey):
    responseCls = KubernetesResponse
    timeout = 60

    def add_default_headers(self, headers):
        """
        Add parameters that are necessary for every request
        If user and password are specified, include a base http auth
        header
        """
        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"

        if self.user_id and self.key:
            auth_string = b("{}:{}".format(self.user_id, self.key))
            user_b64 = base64.b64encode(auth_string)
            headers["Authorization"] = "Basic %s" % (user_b64.decode("utf-8"))
        return headers


class KubernetesDriverMixin:
    """
    Base driver class to be used with various Kubernetes drivers.

    NOTE: This base class can be used in different APIs such as container and
    compute one.
    """

    def __init__(
        self,
        key=None,
        secret=None,
        secure=False,
        host="localhost",
        port=4243,
        key_file=None,
        cert_file=None,
        ca_cert=None,
        ex_token_bearer_auth=False,
    ):
        """
        :param    key: API key or username to be used (required)
        :type     key: ``str``

        :param    secret: Secret password to be used (required)
        :type     secret: ``str``

        :param    secure: Whether to use HTTPS or HTTP. Note: Some providers
                          only support HTTPS, and it is on by default.
        :type     secure: ``bool``

        :param    host: Override hostname used for connections.
        :type     host: ``str``

        :param    port: Override port used for connections.
        :type     port: ``int``

        :param    key_file: Path to the key file used to authenticate (when
                            using key file auth).
        :type     key_file: ``str``

        :param    cert_file: Path to the cert file used to authenticate (when
                             using key file auth).
        :type     cert_file: ``str``

        :param    ex_token_bearer_auth: True to use token bearer auth.
        :type     ex_token_bearer_auth: ``bool``

        :return: ``None``
        """
        if ex_token_bearer_auth:
            self.connectionCls = KubernetesTokenAuthConnection
            if not key:
                msg = 'The token must be a string provided via "key" argument'
                raise ValueError(msg)
            secure = True

        if key_file or cert_file:
            # Certificate based auth is used
            if not (key_file and cert_file):
                raise ValueError("Both key and certificate files are needed")

        if key_file:
            self.connectionCls = KubernetesTLSAuthConnection
            self.key_file = key_file
            self.cert_file = cert_file
            secure = True

        if host and host.startswith("https://"):
            secure = True

        host = self._santize_host(host=host)

        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            key_file=key_file,
            cert_file=cert_file,
        )

        if ca_cert:
            self.connection.connection.ca_cert = ca_cert
        else:
            # do not verify SSL certificate
            warnings.warn(
                "Kubernetes has its own CA, since you didn't supply "
                "a CA certificate be aware that SSL verification "
                "will be disabled for this session."
            )
            self.connection.connection.ca_cert = False

        self.connection.secure = secure
        self.connection.host = host

        if port is not None:
            self.connection.port = port

    def _ex_connection_class_kwargs(self):
        kwargs = {}
        if hasattr(self, "key_file"):
            kwargs["key_file"] = self.key_file
        if hasattr(self, "cert_file"):
            kwargs["cert_file"] = self.cert_file
        return kwargs

    def _santize_host(self, host=None):
        # type: (Optional[str]) -> Optional[str]
        """
        Sanitize "host" argument any remove any protocol prefix (if specified).
        """
        if not host:
            return None

        prefixes = ["http://", "https://"]
        for prefix in prefixes:
            if host.startswith(prefix):
                host = host.lstrip(prefix)

        return host



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib
from libcloud.dns.drivers.vultr import VultrDNSDriver, VultrDNSDriverV2
from libcloud.test.file_fixtures import DNSFileFixtures


class VultrTests(unittest.TestCase):
    def setUp(self):
        VultrMockHttp.type = None
        VultrDNSDriverV2.connectionCls.conn_class = VultrMockHttp
        self.driver = VultrDNSDriver("foo")

    def test_correct_class_is_used(self):
        self.assertIsInstance(self.driver, VultrDNSDriverV2)

    def test_unknown_api_version(self):
        self.assertRaises(NotImplementedError, VultrDNSDriver, "foo", api_version="3")

    def test_list_zones(self):
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 2)
        zone = zones[0]
        self.assertEqual(zone.id, "example.com")
        self.assertEqual(zone.domain, "example.com")
        self.assertEqual(zone.extra["date_created"], "2021-09-07T09:52:18+00:00")

    def test_create_zone(self):
        zone = self.driver.create_zone("example.com")
        self.assertEqual(zone.id, "example.com")
        self.assertEqual(zone.domain, "example.com")
        self.assertEqual(zone.extra["date_created"], "2021-09-07T10:28:34+00:00")

    def test_get_zone(self):
        zone = self.driver.get_zone("example.com")
        self.assertEqual(zone.id, "example.com")
        self.assertEqual(zone.domain, "example.com")
        self.assertEqual(zone.extra["date_created"], "2021-09-07T09:52:18+00:00")

    def test_delete_zone(self):
        zone = self.driver.get_zone("example.com")
        response = self.driver.delete_zone(zone)
        self.assertTrue(response)

    def test_list_records(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone)
        self.assertEqual(len(records), 5)
        record = records[0]
        self.assertEqual(record.id, "123")
        self.assertEqual(record.name, "")
        self.assertEqual(record.type, "NS")
        self.assertEqual(record.data, "ns1.vultr.com")

    def test_create_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.create_record("test1", zone, "A", "192.168.0.11")
        self.assertEqual(record.id, "123")
        self.assertEqual(record.zone.domain, zone.domain)
        self.assertEqual(record.type, "A")
        self.assertEqual(record.name, "test1")
        self.assertEqual(record.extra["priority"], 1)

    def test_update_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone)[0]
        response = self.driver.update_record(
            record, name="test", data="192.168.0.0", extra=dict(ttl=300, priority=1)
        )
        self.assertTrue(response)

    def test_get_record(self):
        zone = self.driver.list_zones()[0]
        temp = self.driver.list_records(zone)[0]
        record = self.driver.get_record(zone.domain, temp.id)
        self.assertEqual(record.id, "123")
        self.assertEqual(record.zone.domain, zone.domain)
        self.assertEqual(record.type, "NS")
        self.assertEqual(record.name, "")
        self.assertEqual(record.extra["priority"], -1)
        self.assertEqual(record.ttl, 300)

    def test_delete_record(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone)[0]
        response = self.driver.delete_record(record)
        self.assertTrue(response)


class VultrMockHttp(MockHttp):
    fixtures = DNSFileFixtures("vultr_v2")

    def _v2_domains(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("list_zones.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        if method == "POST":
            body = self.fixtures.load("create_zone.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_domains_example_com(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("get_zone.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        if method == "DELETE":
            return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])

    def _v2_domains_example_com_records(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("list_records.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        if method == "POST":
            body = self.fixtures.load("create_record.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _v2_domains_example_com_records_123(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("get_record.json")
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])
        elif method == "DELETE":
            return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])
        elif method == "PATCH":
            return (httplib.NO_CONTENT, body, {}, httplib.responses[httplib.NO_CONTENT])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.dns.types import RecordType, ZoneDoesNotExistError, RecordDoesNotExistError
from libcloud.utils.py3 import httplib
from libcloud.common.types import LibcloudError
from libcloud.compute.base import Node
from libcloud.test.secrets import DNS_PARAMS_RACKSPACE
from libcloud.loadbalancer.base import LoadBalancer
from libcloud.test.file_fixtures import DNSFileFixtures
from libcloud.dns.drivers.rackspace import RackspaceDNSDriver, RackspacePTRRecord

# only the 'extra' will be looked at, so pass in minimal data
RDNS_NODE = Node(
    "370b0ff8-3f57-4e10-ac84-e9145ce005841",
    "server1",
    None,
    [],
    [],
    None,
    extra={
        "uri": "https://ord.servers.api.rackspacecloud"
        ".com/v2/905546514/servers/370b0ff8-3f57"
        "-4e10-ac84-e9145ce00584",
        "service_name": "cloudServersOpenStack",
    },
)
RDNS_LB = LoadBalancer(
    "370b0ff8-3f57-4e10-ac84-e9145ce005841",
    "server1",
    None,
    None,
    None,
    None,
    extra={
        "uri": "https://ord.loadbalancers.api."
        "rackspacecloud.com/v2/905546514/"
        "loadbalancers/370b0ff8-3f57-4e10-"
        "ac84-e9145ce00584",
        "service_name": "cloudLoadbalancers",
    },
)


class RackspaceUSTests(unittest.TestCase):
    klass = RackspaceDNSDriver
    endpoint_url = "https://dns.api.rackspacecloud.com/v1.0/11111"
    region = "us"

    def setUp(self):
        self.klass.connectionCls.conn_class = RackspaceMockHttp
        RackspaceMockHttp.type = None

        driver_kwargs = {"region": self.region}
        self.driver = self.klass(*DNS_PARAMS_RACKSPACE, **driver_kwargs)
        self.driver.connection.poll_interval = 0.0
        # normally authentication happens lazily, but we force it here
        self.driver.connection._populate_hosts_and_request_paths()

    def test_force_auth_token_kwargs(self):
        kwargs = {
            "ex_force_auth_token": "some-auth-token",
            "ex_force_base_url": "https://dns.api.rackspacecloud.com/v1.0/11111",
        }
        driver = self.klass(*DNS_PARAMS_RACKSPACE, **kwargs)
        driver.list_zones()

        self.assertEqual(kwargs["ex_force_auth_token"], driver.connection.auth_token)
        self.assertEqual("/v1.0/11111", driver.connection.request_path)

    def test_force_auth_url_kwargs(self):
        kwargs = {
            "ex_force_auth_version": "2.0",
            "ex_force_auth_url": "https://identity.api.rackspace.com",
        }
        driver = self.klass(*DNS_PARAMS_RACKSPACE, **kwargs)

        self.assertEqual(kwargs["ex_force_auth_url"], driver.connection._ex_force_auth_url)
        self.assertEqual(kwargs["ex_force_auth_version"], driver.connection._auth_version)

    def test_gets_auth_2_0_endpoint(self):
        kwargs = {"ex_force_auth_version": "2.0_password", "region": self.region}
        driver = self.klass(*DNS_PARAMS_RACKSPACE, **kwargs)
        driver.connection._populate_hosts_and_request_paths()

        self.assertEqual(self.endpoint_url, driver.connection.get_endpoint())

    def test_list_record_types(self):
        record_types = self.driver.list_record_types()
        self.assertEqual(len(record_types), 8)
        self.assertTrue(RecordType.A in record_types)

    def test_list_zones_success(self):
        zones = self.driver.list_zones()

        self.assertEqual(len(zones), 6)
        self.assertEqual(zones[0].domain, "foo4.bar.com")
        self.assertEqual(zones[0].extra["comment"], "wazaaa")

    def test_list_zones_http_413(self):
        RackspaceMockHttp.type = "413"

        try:
            self.driver.list_zones()
        except LibcloudError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_list_zones_no_results(self):
        RackspaceMockHttp.type = "NO_RESULTS"
        zones = self.driver.list_zones()
        self.assertEqual(len(zones), 0)

    def test_list_records_success(self):
        zone = self.driver.list_zones()[0]
        records = self.driver.list_records(zone=zone)

        self.assertEqual(len(records), 3)
        self.assertEqual(records[0].name, "test3")
        self.assertEqual(records[0].type, RecordType.A)
        self.assertEqual(records[0].data, "127.7.7.7")
        self.assertEqual(records[0].extra["ttl"], 777)
        self.assertEqual(records[0].extra["comment"], "lulz")
        self.assertEqual(records[0].extra["fqdn"], "test3.%s" % (records[0].zone.domain))

    def test_list_records_no_results(self):
        zone = self.driver.list_zones()[0]
        RackspaceMockHttp.type = "NO_RESULTS"
        records = self.driver.list_records(zone=zone)
        self.assertEqual(len(records), 0)

    def test_list_records_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        RackspaceMockHttp.type = "ZONE_DOES_NOT_EXIST"
        try:
            self.driver.list_records(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_get_zone_success(self):
        RackspaceMockHttp.type = "GET_ZONE"
        zone = self.driver.get_zone(zone_id="2946063")

        self.assertEqual(zone.id, "2946063")
        self.assertEqual(zone.domain, "foo4.bar.com")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.extra["email"], "test@test.com")

    def test_get_zone_does_not_exist(self):
        RackspaceMockHttp.type = "DOES_NOT_EXIST"

        try:
            self.driver.get_zone(zone_id="4444")
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, "4444")
        else:
            self.fail("Exception was not thrown")

    def test_get_record_success(self):
        record = self.driver.get_record(zone_id="12345678", record_id="23456789")
        self.assertEqual(record.id, "A-7423034")
        self.assertEqual(record.name, "test3")
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.extra["comment"], "lulz")

    def test_get_record_zone_does_not_exist(self):
        RackspaceMockHttp.type = "ZONE_DOES_NOT_EXIST"

        try:
            self.driver.get_record(zone_id="444", record_id="28536")
        except ZoneDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_get_record_record_does_not_exist(self):
        RackspaceMockHttp.type = "RECORD_DOES_NOT_EXIST"

        try:
            self.driver.get_record(zone_id="12345678", record_id="28536")
        except RecordDoesNotExistError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_zone_success(self):
        RackspaceMockHttp.type = "CREATE_ZONE"

        zone = self.driver.create_zone(
            domain="bar.foo1.com",
            type="master",
            ttl=None,
            extra={"email": "test@test.com"},
        )
        self.assertEqual(zone.id, "2946173")
        self.assertEqual(zone.domain, "bar.foo1.com")
        self.assertEqual(zone.type, "master")
        self.assertEqual(zone.extra["email"], "test@test.com")

    def test_create_zone_validaton_error(self):
        RackspaceMockHttp.type = "CREATE_ZONE_VALIDATION_ERROR"

        try:
            self.driver.create_zone(
                domain="foo.bar.com",
                type="master",
                ttl=10,
                extra={"email": "test@test.com"},
            )
        except Exception as e:
            self.assertEqual(
                str(e),
                "Validation errors: Domain TTL is "
                + "required and must be greater than "
                + "or equal to 300",
            )
        else:
            self.fail("Exception was not thrown")

    def test_update_zone_success(self):
        zone = self.driver.list_zones()[0]
        updated_zone = self.driver.update_zone(zone=zone, extra={"comment": "bar foo"})

        self.assertEqual(zone.extra["comment"], "wazaaa")

        self.assertEqual(updated_zone.id, zone.id)
        self.assertEqual(updated_zone.domain, "foo4.bar.com")
        self.assertEqual(updated_zone.type, zone.type)
        self.assertEqual(updated_zone.ttl, zone.ttl)
        self.assertEqual(updated_zone.extra["comment"], "bar foo")

    def test_update_zone_domain_cannot_be_changed(self):
        zone = self.driver.list_zones()[0]

        try:
            self.driver.update_zone(zone=zone, domain="libcloud.org")
        except LibcloudError:
            pass
        else:
            self.fail("Exception was not thrown")

    def test_create_record_success(self):
        zone = self.driver.list_zones()[0]

        RackspaceMockHttp.type = "CREATE_RECORD"
        record = self.driver.create_record(
            name="www", zone=zone, type=RecordType.A, data="127.1.1.1"
        )

        self.assertEqual(record.id, "A-7423317")
        self.assertEqual(record.name, "www")
        self.assertEqual(record.zone, zone)
        self.assertEqual(record.type, RecordType.A)
        self.assertEqual(record.data, "127.1.1.1")
        self.assertEqual(record.extra["fqdn"], "www.%s" % (zone.domain))

    def test_update_record_success(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        updated_record = self.driver.update_record(record=record, data="127.3.3.3")

        self.assertEqual(record.name, "test3")
        self.assertEqual(record.data, "127.7.7.7")

        self.assertEqual(updated_record.id, record.id)
        self.assertEqual(updated_record.name, record.name)
        self.assertEqual(updated_record.zone, record.zone)
        self.assertEqual(updated_record.type, record.type)
        self.assertEqual(updated_record.data, "127.3.3.3")

    def test_delete_zone_success(self):
        zone = self.driver.list_zones()[0]
        status = self.driver.delete_zone(zone=zone)
        self.assertTrue(status)

    def test_delete_zone_does_not_exist(self):
        zone = self.driver.list_zones()[0]

        RackspaceMockHttp.type = "ZONE_DOES_NOT_EXIST"

        try:
            self.driver.delete_zone(zone=zone)
        except ZoneDoesNotExistError as e:
            self.assertEqual(e.zone_id, zone.id)
        else:
            self.fail("Exception was not thrown")

    def test_delete_record_success(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]
        status = self.driver.delete_record(record=record)
        self.assertTrue(status)

    def test_delete_record_does_not_exist(self):
        zone = self.driver.list_zones()[0]
        record = self.driver.list_records(zone=zone)[0]

        RackspaceMockHttp.type = "RECORD_DOES_NOT_EXIST"

        try:
            self.driver.delete_record(record=record)
        except RecordDoesNotExistError as e:
            self.assertEqual(e.record_id, record.id)
        else:
            self.fail("Exception was not thrown")

    def test_to_full_record_name_name_provided(self):
        domain = "foo.bar"
        name = "test"
        self.assertEqual(self.driver._to_full_record_name(domain, name), "test.foo.bar")

    def test_to_full_record_name_name_not_provided(self):
        domain = "foo.bar"
        name = None
        self.assertEqual(self.driver._to_full_record_name(domain, name), "foo.bar")

    def test_to_partial_record_name(self):
        domain = "example.com"
        names = [
            "test.example.com",
            "foo.bar.example.com",
            "example.com.example.com",
            "example.com",
        ]
        expected_values = ["test", "foo.bar", "example.com", None]

        for name, expected_value in zip(names, expected_values):
            value = self.driver._to_partial_record_name(domain=domain, name=name)
            self.assertEqual(value, expected_value)

    def test_ex_create_ptr_success(self):
        ip = "127.1.1.1"
        domain = "www.foo4.bar.com"
        record = self.driver.ex_create_ptr_record(RDNS_NODE, ip, domain)
        self.assertEqual(record.ip, ip)
        self.assertEqual(record.domain, domain)
        self.assertEqual(record.extra["uri"], RDNS_NODE.extra["uri"])
        self.assertEqual(record.extra["service_name"], RDNS_NODE.extra["service_name"])

        self.driver.ex_create_ptr_record(RDNS_LB, ip, domain)

    def test_ex_list_ptr_success(self):
        records = self.driver.ex_iterate_ptr_records(RDNS_NODE)
        for record in records:
            self.assertTrue(isinstance(record, RackspacePTRRecord))
            self.assertEqual(record.type, RecordType.PTR)
            self.assertEqual(record.extra["uri"], RDNS_NODE.extra["uri"])
            self.assertEqual(record.extra["service_name"], RDNS_NODE.extra["service_name"])

    def test_ex_list_ptr_not_found(self):
        RackspaceMockHttp.type = "RECORD_DOES_NOT_EXIST"

        try:
            records = self.driver.ex_iterate_ptr_records(RDNS_NODE)
        except Exception as exc:
            self.fail("PTR Records list 404 threw %s" % exc)

        try:
            next(records)
            self.fail("PTR Records list 404 did not produce an empty list")
        except StopIteration:
            self.assertTrue(True, "Got empty list on 404")

    def text_ex_get_ptr_success(self):
        service_name = "cloudServersOpenStack"
        records = self.driver.ex_iterate_ptr_records(service_name)
        original = next(records)
        found = self.driver.ex_get_ptr_record(service_name, original.id)
        for attr in dir(original):
            self.assertEqual(getattr(found, attr), getattr(original, attr))

    def text_update_ptr_success(self):
        records = self.driver.ex_iterate_ptr_records(RDNS_NODE)
        original = next(records)

        updated = self.driver.ex_update_ptr_record(original, domain=original.domain)
        self.assertEqual(original.id, updated.id)

        extra_update = {"ttl": original.extra["ttl"]}
        updated = self.driver.ex_update_ptr_record(original, extra=extra_update)
        self.assertEqual(original.id, updated.id)

        updated = self.driver.ex_update_ptr_record(original, "new-domain")
        self.assertEqual(original.id, updated.id)

    def test_ex_delete_ptr_success(self):
        records = self.driver.ex_iterate_ptr_records(RDNS_NODE)
        original = next(records)
        self.assertTrue(self.driver.ex_delete_ptr_record(original))


class RackspaceUKTests(RackspaceUSTests):
    klass = RackspaceDNSDriver
    endpoint_url = "https://lon.dns.api.rackspacecloud.com/v1.0/11111"
    region = "uk"


class RackspaceMockHttp(MockHttp):
    fixtures = DNSFileFixtures("rackspace")
    base_headers = {"content-type": "application/json"}

    def _v2_0_tokens(self, method, url, body, headers):
        body = self.fixtures.load("auth_2_0.json")
        headers = {"content-type": "application/json"}
        return (httplib.OK, body, headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains(self, method, url, body, headers):
        body = self.fixtures.load("list_zones_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_413(self, method, url, body, headers):
        body = ""
        return (
            httplib.REQUEST_ENTITY_TOO_LARGE,
            body,
            self.base_headers,
            httplib.responses[httplib.REQUEST_ENTITY_TOO_LARGE],
        )

    def _v1_0_11111_domains_NO_RESULTS(self, method, url, body, headers):
        body = self.fixtures.load("list_zones_no_results.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_2946063(self, method, url, body, headers):
        if method == "GET":
            body = self.fixtures.load("list_records_success.json")
        elif method == "PUT":
            # Async - update_zone
            body = self.fixtures.load("update_zone_success.json")
        elif method == "DELETE":
            # Async - delete_zone
            body = self.fixtures.load("delete_zone_success.json")

        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_2946063_NO_RESULTS(self, method, url, body, headers):
        body = self.fixtures.load("list_records_no_results.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_2946063_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("does_not_exist.json")
        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.NOT_FOUND],
        )

    def _v1_0_11111_domains_2946063_GET_ZONE(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_4444_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("does_not_exist.json")
        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.NOT_FOUND],
        )

    def _v1_0_11111_domains_12345678(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_12345678_records_23456789(self, method, url, body, headers):
        body = self.fixtures.load("get_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_444_ZONE_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("does_not_exist.json")
        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.NOT_FOUND],
        )

    def _v1_0_11111_domains_12345678_RECORD_DOES_NOT_EXIST(self, method, url, body, headers):
        body = self.fixtures.load("get_zone_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_12345678_records_28536_RECORD_DOES_NOT_EXIST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("does_not_exist.json")
        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.NOT_FOUND],
        )

    def _v1_0_11111_domains_CREATE_ZONE(self, method, url, body, headers):
        # Async response - create_zone
        body = self.fixtures.load("create_zone_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_status_288795f9_e74d_48be_880b_a9e36e0de61e_CREATE_ZONE(
        self, method, url, body, headers
    ):
        # Async status - create_zone
        body = self.fixtures.load("create_zone_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_CREATE_ZONE_VALIDATION_ERROR(self, method, url, body, headers):
        body = self.fixtures.load("create_zone_validation_error.json")
        return (
            httplib.BAD_REQUEST,
            body,
            self.base_headers,
            httplib.responses[httplib.BAD_REQUEST],
        )

    def _v1_0_11111_status_116a8f17_38ac_4862_827c_506cd04800d5(self, method, url, body, headers):
        # Aync status - update_zone
        body = self.fixtures.load("update_zone_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_status_586605c8_5739_43fb_8939_f3a2c4c0e99c_CREATE_RECORD(
        self, method, url, body, headers
    ):
        # Aync status - create_record
        body = self.fixtures.load("create_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_2946063_records_CREATE_RECORD(self, method, url, body, headers):
        # Aync response - create_record
        body = self.fixtures.load("create_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_status_251c0d0c_95bc_4e09_b99f_4b8748b66246(self, method, url, body, headers):
        # Aync response - update_record
        body = self.fixtures.load("update_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_2946063_records_A_7423034(self, method, url, body, headers):
        # Aync response - update_record
        body = self.fixtures.load("update_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_status_0b40cd14_2e5d_490f_bb6e_fdc65d1118a9(self, method, url, body, headers):
        # Async status - delete_zone
        body = self.fixtures.load("delete_zone_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_status_0b40cd14_2e5d_490f_bb6e_fdc65d1118a9_RECORD_DOES_NOT_EXIST(
        self, method, url, body, headers
    ):
        # Async status - delete_record
        body = self.fixtures.load("delete_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_domains_2946063_records_A_7423034_RECORD_DOES_NOT_EXIST(
        self, method, url, body, headers
    ):
        # Async response - delete_record
        body = self.fixtures.load("does_not_exist.json")
        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.NOT_FOUND],
        )

    def _v1_0_11111_rdns_cloudServersOpenStack(self, method, url, body, headers):
        if method == "DELETE":
            body = self.fixtures.load("delete_ptr_record_success.json")
            return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])
        else:
            body = self.fixtures.load("list_ptr_records_success.json")
            return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_rdns_cloudServersOpenStack_RECORD_DOES_NOT_EXIST(
        self, method, url, body, headers
    ):
        body = self.fixtures.load("does_not_exist.json")
        return (
            httplib.NOT_FOUND,
            body,
            self.base_headers,
            httplib.responses[httplib.NOT_FOUND],
        )

    def _v1_0_11111_rdns_cloudServersOpenStack_PTR_7423034(self, method, url, body, headers):
        body = self.fixtures.load("get_ptr_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_rdns(self, method, url, body, headers):
        body = self.fixtures.load("create_ptr_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_status_12345678_5739_43fb_8939_f3a2c4c0e99c(self, method, url, body, headers):
        body = self.fixtures.load("create_ptr_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])

    def _v1_0_11111_status_12345678_2e5d_490f_bb6e_fdc65d1118a9(self, method, url, body, headers):
        body = self.fixtures.load("delete_ptr_record_success.json")
        return (httplib.OK, body, self.base_headers, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from unittest import mock

from libcloud.test import unittest
from libcloud.compute.drivers.libvirt_driver import LibvirtNodeDriver, have_libvirt

__all__ = ["LibvirtNodeDriverTestCase"]


@unittest.skipIf(not have_libvirt, "libvirt not available, skipping tests")
@mock.patch("libcloud.compute.drivers.libvirt_driver.libvirt", autospec=True)
class LibvirtNodeDriverTestCase(unittest.TestCase):
    arp_output_str = """? (1.2.10.80) at 52:54:00:bc:f9:6c [ether] on br0
? (1.2.10.33) at 52:54:00:04:89:51 [ether] on br0
? (1.2.10.97) at 52:54:00:c6:40:ec [ether] on br0
? (1.2.10.40) at 52:54:00:77:1c:83 [ether] on br0"""
    ip_output_str = """1.2.10.80 dev br0 lladdr 52:54:00:bc:f9:6c STALE
1.2.10.33 dev br0 lladdr 52:54:00:04:89:51 REACHABLE
1.2.10.97 dev br0 lladdr 52:54:00:c6:40:ec DELAY
1.2.10.40 dev br0 lladdr 52:54:00:77:1c:83 STALE"""
    bad_output_str = """1.2.10.80 dev br0  52:54:00:bc:f9:6c STALE
1.2.10.33 dev br0 lladdr 52:54:00:04:89:51 REACHABLE
1.2.10.97 dev br0 lladdr
1.2.10.40 dev br0 lladdr 52:54:00:77:1c:83 STALE"""

    from libcloud.utils.py3 import b

    arp_output_str = b(arp_output_str)
    ip_output_str = b(ip_output_str)
    bad_output_str = b(bad_output_str)

    def _assert_arp_table(self, arp_table):
        self.assertIn("52:54:00:bc:f9:6c", arp_table)
        self.assertIn("52:54:00:04:89:51", arp_table)
        self.assertIn("52:54:00:c6:40:ec", arp_table)
        self.assertIn("52:54:00:77:1c:83", arp_table)
        self.assertIn("1.2.10.80", arp_table["52:54:00:bc:f9:6c"])
        self.assertIn("1.2.10.33", arp_table["52:54:00:04:89:51"])
        self.assertIn("1.2.10.97", arp_table["52:54:00:c6:40:ec"])
        self.assertIn("1.2.10.40", arp_table["52:54:00:77:1c:83"])

    def test_arp_map(self, *args, **keywargs):
        driver = LibvirtNodeDriver("")
        arp_table = driver._parse_ip_table_arp(self.arp_output_str)
        self._assert_arp_table(arp_table)

    def test_ip_map(self, *args, **keywargs):
        driver = LibvirtNodeDriver("")
        arp_table = driver._parse_ip_table_neigh(self.ip_output_str)
        self._assert_arp_table(arp_table)

    def test_bad_map(self, *args, **keywargs):
        driver = LibvirtNodeDriver("")
        arp_table = driver._parse_ip_table_neigh(self.bad_output_str)
        # we should at least get the correctly formatted lines
        self.assertEqual(len(arp_table), 2)
        arp_table = driver._parse_ip_table_neigh(self.arp_output_str)
        # nothing should match if the wrong output is sent
        self.assertEqual(len(arp_table), 0)

    def test_list_nodes(self, *args, **keywargs):
        driver = LibvirtNodeDriver("")
        nodes = driver.list_nodes()
        self.assertEqual(type([]), type(nodes))
        self.assertEqual(len(nodes), 0)


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib
from libcloud.test.secrets import LB_ALB_PARAMS
from libcloud.loadbalancer.base import Member
from libcloud.loadbalancer.types import State
from libcloud.test.file_fixtures import LoadBalancerFileFixtures
from libcloud.loadbalancer.drivers.alb import ApplicationLBDriver


class ApplicationLBTests(unittest.TestCase):
    # defaults from fixtures
    balancer_id = "arn:aws:elasticloadbalancing:us-east-1:111111111111:loadbalancer/app/Test-ALB/1111111111111111"
    target_group_id = "arn:aws:elasticloadbalancing:us-east-1:111111111111:targetgroup/Test-ALB-tg/1111111111111111"
    listener_id = "arn:aws:elasticloadbalancing:us-east-1:111111111111:listener/app/Test-ALB/1111111111111111/1111111111111111"
    rule_id = (
        "arn:aws:elasticloadbalancing:us-east-1:111111111111:listener-rule/app/Test-Develop-App-LB/1111111111111111/"
        "1111111111111111/1111111111111111"
    )
    ssl_cert_id = "arn:aws:iam::111111111111:server-certificate/test.certificate"

    def setUp(self):
        ApplicationLBMockHttp.test = self
        ApplicationLBDriver.connectionCls.conn_class = ApplicationLBMockHttp
        ApplicationLBMockHttp.type = None
        ApplicationLBMockHttp.use_param = "Action"
        self.driver = ApplicationLBDriver(*LB_ALB_PARAMS)

    def test_instantiate_driver_with_token(self):
        token = "temporary_credentials_token"
        driver = ApplicationLBDriver(*LB_ALB_PARAMS, **{"token": token})
        self.assertTrue(hasattr(driver, "token"), "Driver has no attribute token")
        self.assertEqual(token, driver.token, "Driver token does not match with provided token")

    def test_driver_with_token_signature_version(self):
        token = "temporary_credentials_token"
        driver = ApplicationLBDriver(*LB_ALB_PARAMS, **{"token": token})
        kwargs = driver._ex_connection_class_kwargs()
        self.assertTrue(
            ("signature_version" in kwargs), "Driver has no attribute signature_version"
        )
        self.assertEqual(
            "4",
            kwargs["signature_version"],
            "Signature version is not 4 with temporary credentials",
        )

    def test_list_protocols(self):
        protocols = self.driver.list_protocols()
        self.assertEqual(len(protocols), 2)
        self.assertTrue("http" in protocols)
        self.assertTrue("https" in protocols)

    def test_list_balancers(self):
        balancers = self.driver.list_balancers()
        self.assertEqual(len(balancers), 1)
        self.assertEqual(balancers[0].id, self.balancer_id)
        self.assertEqual(balancers[0].name, "Test-ALB")

    def test_get_balancer(self):
        balancer = self.driver.get_balancer(balancer_id=self.balancer_id)
        self.assertEqual(balancer.id, self.balancer_id)
        self.assertEqual(balancer.extra["listeners"][0].balancer.id, self.balancer_id)
        self.assertEqual(balancer.name, "Test-ALB")
        self.assertEqual(balancer.state, State.UNKNOWN)

    def test_create_balancer(self):
        balancer = self.driver.create_balancer(
            name="Test-ALB",
            port=443,
            protocol="HTTPS",
            algorithm=None,
            members=[Member(id="i-01111111111111111", ip=None, port=443)],
            ex_scheme="internet-facing",
            ex_security_groups=["sg-11111111"],
            ex_subnets=["subnet-11111111", "subnet-22222222"],
            ex_tags={},
            ex_ssl_cert_arn=self.ssl_cert_id,
        )
        self.assertEqual(balancer.id, self.balancer_id)
        self.assertEqual(balancer.name, "Test-ALB")
        self.assertEqual(balancer.state, State.UNKNOWN)
        self.assertEqual(balancer.port, 443)

    def test_ex_create_balancer(self):
        balancer = self.driver.ex_create_balancer(
            name="Test-ALB",
            addr_type="ipv4",
            scheme="internet-facing",
            security_groups=["sg-11111111"],
            subnets=["subnet-11111111", "subnet-22222222"],
        )
        self.assertEqual(balancer.id, self.balancer_id)
        self.assertEqual(balancer.name, "Test-ALB")
        self.assertEqual(balancer.state, State.UNKNOWN)

    def test_ex_create_target_group(self):
        target_group = self.driver.ex_create_target_group(
            name="Test-ALB-tg",
            port=443,
            proto="HTTPS",
            vpc="vpc-11111111",
            health_check_interval=30,
            health_check_path="/",
            health_check_port="traffic-port",
            health_check_proto="HTTP",
            health_check_timeout=5,
            health_check_matcher="200",
            healthy_threshold=5,
            unhealthy_threshold=2,
        )
        self.assertTrue(hasattr(target_group, "id"), 'Target group is missing "id" field')
        self.assertTrue(hasattr(target_group, "members"), 'Target group is missing "members" field')
        self.assertEqual(target_group.name, "Test-ALB-tg")
        self.assertEqual(target_group.port, 443)
        self.assertEqual(target_group.protocol, "HTTPS")
        self.assertEqual(target_group.vpc, "vpc-11111111")
        self.assertEqual(target_group.health_check_timeout, 5)
        self.assertEqual(target_group.health_check_port, "traffic-port")
        self.assertEqual(target_group.health_check_path, "/")
        self.assertEqual(target_group.health_check_matcher, "200")
        self.assertEqual(target_group.health_check_proto, "HTTPS")
        self.assertEqual(target_group.health_check_interval, 30)
        self.assertEqual(target_group.healthy_threshold, 5)
        self.assertEqual(target_group.unhealthy_threshold, 2)

    def test_ex_register_targets(self):
        self.driver.get_balancer(self.balancer_id)
        target_group = self.driver.ex_get_target_group(self.target_group_id)
        members = [Member("i-01111111111111111", "10.0.0.0", 443)]
        targets_not_registered = self.driver.ex_register_targets(
            target_group=target_group, members=members
        )
        self.assertTrue(
            targets_not_registered,
            "ex_register_targets is expected to return True on success",
        )

    def test_ex_create_listener(self):
        balancer = self.driver.get_balancer(self.balancer_id)
        target_group = self.driver.ex_get_target_group(self.target_group_id)
        listener = self.driver.ex_create_listener(
            balancer=balancer,
            port=443,
            proto="HTTPS",
            target_group=target_group,
            action="forward",
            ssl_cert_arn=self.ssl_cert_id,
            ssl_policy="ELBSecurityPolicy-2016-08",
        )

        self.assertTrue(hasattr(listener, "id"), 'Listener is missing "id" field')
        self.assertTrue(hasattr(listener, "rules"), 'Listener is missing "rules" field')
        self.assertTrue(hasattr(listener, "balancer"), 'Listener is missing "balancer" field')
        self.assertEqual(listener.balancer.id, balancer.id)
        self.assertEqual(listener.rules[0].target_group.id, target_group.id)
        self.assertEqual(listener.port, 443)
        self.assertEqual(listener.protocol, "HTTPS")
        self.assertEqual(listener.action, "forward")
        self.assertEqual(listener.ssl_certificate, self.ssl_cert_id)
        self.assertEqual(listener.ssl_policy, "ELBSecurityPolicy-2016-08")

    def test_ex_create_rule(self):
        balancer = self.driver.get_balancer(self.balancer_id)
        listener = balancer.extra.get("listeners")[0]
        target_group = self.driver.ex_get_target_group(self.target_group_id)
        rule = self.driver.ex_create_listener_rule(
            listener=listener,
            priority=10,
            target_group=target_group,
            action="forward",
            condition_field="path-pattern",
            condition_value="/img/*",
        )

        self.assertTrue(hasattr(rule, "id"), 'Rule is missing "id" field')
        self.assertTrue(hasattr(rule, "conditions"), 'Rule is missing "conditions" field')
        self.assertEqual(rule.is_default, False)
        self.assertEqual(rule.priority, "10")
        self.assertEqual(rule.target_group.id, self.target_group_id)
        self.assertTrue(
            ("/img/*" in rule.conditions["path-pattern"]),
            "Rule is missing test condition",
        )

    def test_ex_get_balancer_tags(self):
        balancer = self.driver.get_balancer(balancer_id=self.balancer_id)
        self.assertTrue(("tags" in balancer.extra), "No tags dict found in balancer.extra")
        tags = self.driver._ex_get_balancer_tags(balancer)
        self.assertEqual(tags["project"], "lima")

    def test_ex_get_target_group(self):
        target_group = self.driver.ex_get_target_group(self.target_group_id)
        target_group_fields = (
            "id",
            "name",
            "protocol",
            "port",
            "vpc",
            "health_check_timeout",
            "health_check_port",
            "health_check_path",
            "health_check_proto",
            "health_check_matcher",
            "health_check_interval",
            "healthy_threshold",
            "unhealthy_threshold",
            "_balancers",
            "_balancers_arns",
            "_members",
            "_driver",
        )

        for field in target_group_fields:
            self.assertTrue(
                (field in target_group.__dict__),
                "Field [%s] is missing in ALBTargetGroup object" % field,
            )

        self.assertEqual(target_group.id, self.target_group_id)

    def test_ex_get_listener(self):
        listener = self.driver.ex_get_listener(self.listener_id)
        listener_fields = (
            "id",
            "protocol",
            "port",
            "action",
            "ssl_policy",
            "ssl_certificate",
            "_balancer",
            "_balancer_arn",
            "_rules",
            "_driver",
        )

        for field in listener_fields:
            self.assertTrue(
                (field in listener.__dict__),
                "Field [%s] is missing in ALBListener object" % field,
            )

        self.assertEqual(listener.id, self.listener_id)

    def test_ex_get_rule(self):
        rule = self.driver.ex_get_rule(self.rule_id)
        rule_fields = (
            "id",
            "is_default",
            "priority",
            "conditions",
            "_listener",
            "_listener_arn",
            "_target_group",
            "_target_group_arn",
            "_driver",
        )

        for field in rule_fields:
            self.assertTrue(
                (field in rule.__dict__),
                "Field [%s] is missing in ALBRule object" % field,
            )

        self.assertEqual(rule.id, self.rule_id)

    def test_ex_get_target_group_members(self):
        target_group = self.driver.ex_get_target_group(self.target_group_id)
        target_group_members = self.driver._ex_get_target_group_members(target_group)
        self.assertEqual(len(target_group_members), 1)
        self.assertTrue(
            hasattr(target_group_members[0], "id"),
            'Target group member is missing "id" field',
        )
        self.assertTrue(
            hasattr(target_group_members[0], "port"),
            'Target group member is missing "port" field',
        )
        self.assertTrue(
            ("health" in target_group_members[0].extra),
            'Target group member is missing "health" field',
        )

    def test_ex_get_balancer_listeners(self):
        balancer = self.driver.get_balancer(balancer_id=self.balancer_id)
        listeners = self.driver._ex_get_balancer_listeners(balancer)
        self.assertEqual(len(listeners), 1)
        self.assertTrue(hasattr(listeners[0], "id"), 'Listener is missing "id" field')
        self.assertTrue(hasattr(listeners[0], "port"), 'Listener is missing "port" field')
        self.assertTrue(hasattr(listeners[0], "protocol"), 'Listener is missing "protocol" field')
        self.assertTrue(hasattr(listeners[0], "rules"), 'Listener is missing "rules" field')

    def test_ex_get_rules_for_listener(self):
        listener = self.driver.ex_get_listener(self.listener_id)
        listener_rules = self.driver._ex_get_rules_for_listener(listener)
        self.assertEqual(len(listener_rules), 1)
        self.assertTrue(hasattr(listener_rules[0], "id"), 'Rule is missing "id" field')
        self.assertTrue(hasattr(listener_rules[0], "is_default"), 'Rule is missing "port" field')
        self.assertTrue(hasattr(listener_rules[0], "priority"), 'Rule is missing "priority" field')
        self.assertTrue(
            hasattr(listener_rules[0], "target_group"),
            'Rule is missing "target_group" field',
        )
        self.assertTrue(
            hasattr(listener_rules[0], "conditions"),
            'Rule is missing "conditions" field',
        )


# Commented out to avoid confusion. In AWS ALB relation between load balancer
# and target group/members is indirect. So it's better to go through full chain
# to obtain required object(s).
# Chain is: balancer->listener->rule->target group->member
#
#    def test_balancer_list_members(self):
#         balancer = self.driver.get_balancer(balancer_id=self.balancer_id)
#         members = balancer.list_members()
#         self.assertEqual(len(members), 1)
#         self.assertEqual(members[0].balancer, balancer)
#         self.assertEqual('i-01111111111111111', members[0].id)
#
#    def test_ex_get_balancer_target_groups(self):
#         balancer = self.driver.get_balancer(balancer_id=self.balancer_id)
#         target_groups = self.driver._ex_get_balancer_target_groups(balancer)
#         self.assertEqual(len(target_groups), 1)
#         self.assertTrue(hasattr(target_groups[0], 'id'), 'Target group is missing "id" field')
#         self.assertTrue(hasattr(target_groups[0], 'name'), 'Target group is missing "port" field')
#         self.assertTrue(hasattr(target_groups[0], 'members'), 'Target group is missing "members" field')


class ApplicationLBMockHttp(MockHttp):
    fixtures = LoadBalancerFileFixtures("alb")

    def _2015_12_01_DescribeLoadBalancers(self, method, url, body, headers):
        body = self.fixtures.load("describe_load_balancers.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_DescribeListeners(self, method, url, body, headers):
        body = self.fixtures.load("describe_load_balancer_listeters.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_DescribeRules(self, method, url, body, headers):
        body = self.fixtures.load("describe_load_balancer_rules.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_DescribeTargetGroups(self, method, url, body, headers):
        body = self.fixtures.load("describe_load_balancer_target_groups.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_DescribeTargetHealth(self, method, url, body, headers):
        body = self.fixtures.load("describe_target_health.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_DescribeTags(self, method, url, body, headers):
        body = self.fixtures.load("describe_tags.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_CreateLoadBalancer(self, method, url, body, headers):
        body = self.fixtures.load("create_balancer.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_CreateTargetGroup(self, method, url, body, headers):
        body = self.fixtures.load("create_target_group.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_CreateListener(self, method, url, body, headers):
        body = self.fixtures.load("create_listener.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_CreateRule(self, method, url, body, headers):
        body = self.fixtures.load("create_rule.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _2015_12_01_RegisterTargets(self, method, url, body, headers):
        body = self.fixtures.load("register_targets.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import ModuleType
from typing import TYPE_CHECKING, Type, Union

from libcloud.dns.types import OLD_CONSTANT_TO_NEW_MAPPING, Provider
from libcloud.common.providers import get_driver as _get_provider_driver
from libcloud.common.providers import set_driver as _set_provider_driver

if TYPE_CHECKING:
    # NOTE: This is needed to avoid having setup.py depend on requests
    from libcloud.dns.base import DNSDriver


__all__ = ["DRIVERS", "get_driver", "set_driver"]

DRIVERS = {
    Provider.DUMMY: ("libcloud.dns.drivers.dummy", "DummyDNSDriver"),
    Provider.LINODE: ("libcloud.dns.drivers.linode", "LinodeDNSDriver"),
    Provider.ZERIGO: ("libcloud.dns.drivers.zerigo", "ZerigoDNSDriver"),
    Provider.RACKSPACE: ("libcloud.dns.drivers.rackspace", "RackspaceDNSDriver"),
    Provider.ROUTE53: ("libcloud.dns.drivers.route53", "Route53DNSDriver"),
    Provider.GANDI: ("libcloud.dns.drivers.gandi", "GandiDNSDriver"),
    Provider.GANDI_LIVE: ("libcloud.dns.drivers.gandi_live", "GandiLiveDNSDriver"),
    Provider.GOOGLE: ("libcloud.dns.drivers.google", "GoogleDNSDriver"),
    Provider.DIGITAL_OCEAN: (
        "libcloud.dns.drivers.digitalocean",
        "DigitalOceanDNSDriver",
    ),
    Provider.WORLDWIDEDNS: ("libcloud.dns.drivers.worldwidedns", "WorldWideDNSDriver"),
    Provider.DNSIMPLE: ("libcloud.dns.drivers.dnsimple", "DNSimpleDNSDriver"),
    Provider.POINTDNS: ("libcloud.dns.drivers.pointdns", "PointDNSDriver"),
    Provider.VULTR: ("libcloud.dns.drivers.vultr", "VultrDNSDriver"),
    Provider.LIQUIDWEB: ("libcloud.dns.drivers.liquidweb", "LiquidWebDNSDriver"),
    Provider.ZONOMI: ("libcloud.dns.drivers.zonomi", "ZonomiDNSDriver"),
    Provider.DURABLEDNS: ("libcloud.dns.drivers.durabledns", "DurableDNSDriver"),
    Provider.AURORADNS: ("libcloud.dns.drivers.auroradns", "AuroraDNSDriver"),
    Provider.GODADDY: ("libcloud.dns.drivers.godaddy", "GoDaddyDNSDriver"),
    Provider.CLOUDFLARE: ("libcloud.dns.drivers.cloudflare", "CloudFlareDNSDriver"),
    Provider.NFSN: ("libcloud.dns.drivers.nfsn", "NFSNDNSDriver"),
    Provider.NSONE: ("libcloud.dns.drivers.nsone", "NsOneDNSDriver"),
    Provider.LUADNS: ("libcloud.dns.drivers.luadns", "LuadnsDNSDriver"),
    Provider.BUDDYNS: ("libcloud.dns.drivers.buddyns", "BuddyNSDNSDriver"),
    Provider.POWERDNS: ("libcloud.dns.drivers.powerdns", "PowerDNSDriver"),
    Provider.ONAPP: ("libcloud.dns.drivers.onapp", "OnAppDNSDriver"),
    Provider.RCODEZERO: ("libcloud.dns.drivers.rcodezero", "RcodeZeroDNSDriver"),
    # Deprecated
    Provider.RACKSPACE_US: ("libcloud.dns.drivers.rackspace", "RackspaceUSDNSDriver"),
    Provider.RACKSPACE_UK: ("libcloud.dns.drivers.rackspace", "RackspaceUKDNSDriver"),
}


def get_driver(provider):
    # type: (Union[Provider, str]) -> Type[DNSDriver]
    deprecated_constants = OLD_CONSTANT_TO_NEW_MAPPING
    return _get_provider_driver(
        drivers=DRIVERS, provider=provider, deprecated_constants=deprecated_constants
    )


def set_driver(provider, module, klass):
    # type: (Union[Provider, str], ModuleType, type) -> Type[DNSDriver]
    return _set_provider_driver(drivers=DRIVERS, provider=provider, module=module, klass=klass)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib
from libcloud.common.base import Response, Connection, XmlResponse, JsonResponse
from libcloud.test.file_fixtures import ComputeFileFixtures


class FileFixturesTests(unittest.TestCase):
    def test_success(self):
        f = ComputeFileFixtures("meta")
        self.assertEqual("Hello, World!", f.load("helloworld.txt"))

    def test_failure(self):
        f = ComputeFileFixtures("meta")
        self.assertRaises(IOError, f.load, "nil")

    def test_unicode(self):
        f = ComputeFileFixtures("meta")
        self.assertEqual("Ś", f.load("unicode.txt"))


class MockHttpFileFixturesTests(unittest.TestCase):
    """
    Test the behaviour of MockHttp
    """

    def setUp(self):
        Connection.conn_class = TestMockHttp
        Connection.responseCls = Response
        self.connection = Connection()

    def test_unicode_response(self):
        r = self.connection.request("/unicode")
        self.assertEqual(r.parse_body(), "Ś")

    def test_json_unicode_response(self):
        self.connection.responseCls = JsonResponse
        r = self.connection.request("/unicode/json")
        self.assertEqual(r.object, {"test": "Ś"})

    def test_xml_unicode_response(self):
        self.connection.responseCls = XmlResponse
        response = self.connection.request("/unicode/xml")
        self.assertEqual(response.object.text, "Ś")


class TestMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("meta")

    def _unicode(self, method, url, body, headers):
        body = self.fixtures.load("unicode.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _unicode_json(self, method, url, body, headers):
        body = self.fixtures.load("unicode.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _unicode_xml(self, method, url, body, headers):
        body = self.fixtures.load("unicode.xml")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _ascii(self, method, url, body, headers):
        body = self.fixtures.load("helloworld.txt")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib
from libcloud.compute.base import Node
from libcloud.test.compute import TestCaseMixin
from libcloud.test.secrets import VPSNET_PARAMS
from libcloud.compute.types import NodeState
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.vpsnet import VPSNetNodeDriver


class VPSNetTests(unittest.TestCase, TestCaseMixin):
    def setUp(self):
        VPSNetNodeDriver.connectionCls.conn_class = VPSNetMockHttp
        self.driver = VPSNetNodeDriver(*VPSNET_PARAMS)

    def test_create_node(self):
        VPSNetMockHttp.type = "create"
        image = self.driver.list_images()[0]
        size = self.driver.list_sizes()[0]
        node = self.driver.create_node("foo", image, size)
        self.assertEqual(node.name, "foo")

    def test_list_nodes(self):
        VPSNetMockHttp.type = "virtual_machines"
        node = self.driver.list_nodes()[0]
        self.assertEqual(node.id, "1384")
        self.assertEqual(node.state, NodeState.RUNNING)

    def test_reboot_node(self):
        VPSNetMockHttp.type = "virtual_machines"
        node = self.driver.list_nodes()[0]

        VPSNetMockHttp.type = "reboot"
        ret = self.driver.reboot_node(node)
        self.assertEqual(ret, True)

    def test_destroy_node(self):
        VPSNetMockHttp.type = "delete"
        node = Node("2222", None, None, None, None, self.driver)
        ret = self.driver.destroy_node(node)
        self.assertTrue(ret)
        VPSNetMockHttp.type = "delete_fail"
        node = Node("2223", None, None, None, None, self.driver)
        self.assertRaises(Exception, self.driver.destroy_node, node)

    def test_list_images(self):
        VPSNetMockHttp.type = "templates"
        ret = self.driver.list_images()
        self.assertEqual(ret[0].id, "9")
        self.assertEqual(ret[-1].id, "160")

    def test_list_sizes(self):
        VPSNetMockHttp.type = "sizes"
        ret = self.driver.list_sizes()
        self.assertEqual(len(ret), 1)
        self.assertEqual(ret[0].id, "1")
        self.assertEqual(ret[0].name, "1 Node")

    def test_destroy_node_response(self):
        # should return a node object
        node = Node("2222", None, None, None, None, self.driver)
        VPSNetMockHttp.type = "delete"
        ret = self.driver.destroy_node(node)
        self.assertTrue(isinstance(ret, bool))

    def test_reboot_node_response(self):
        # should return a node object
        VPSNetMockHttp.type = "virtual_machines"
        node = self.driver.list_nodes()[0]
        VPSNetMockHttp.type = "reboot"
        ret = self.driver.reboot_node(node)
        self.assertTrue(isinstance(ret, bool))


class VPSNetMockHttp(MockHttp):
    fixtures = ComputeFileFixtures("vpsnet")

    def _nodes_api10json_sizes(self, method, url, body, headers):
        body = """[{"slice":{"virtual_machine_id":8592,"id":12256,"consumer_id":0}},
                   {"slice":{"virtual_machine_id":null,"id":12258,"consumer_id":0}},
                   {"slice":{"virtual_machine_id":null,"id":12434,"consumer_id":0}}]"""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _nodes_api10json_create(self, method, url, body, headers):
        body = """[{"slice":{"virtual_machine_id":8592,"id":12256,"consumer_id":0}},
                   {"slice":{"virtual_machine_id":null,"id":12258,"consumer_id":0}},
                   {"slice":{"virtual_machine_id":null,"id":12434,"consumer_id":0}}]"""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _virtual_machines_2222_api10json_delete_fail(self, method, url, body, headers):
        return (httplib.FORBIDDEN, "", {}, httplib.responses[httplib.FORBIDDEN])

    def _virtual_machines_2222_api10json_delete(self, method, url, body, headers):
        return (httplib.OK, "", {}, httplib.responses[httplib.OK])

    def _virtual_machines_1384_reboot_api10json_reboot(self, method, url, body, headers):
        body = """{
              "virtual_machine":
                {
                  "running": true,
                  "updated_at": "2009-05-15T06:55:02-04:00",
                  "power_action_pending": false,
                  "system_template_id": 41,
                  "id": 1384,
                  "cloud_id": 3,
                  "domain_name": "demodomain.com",
                  "hostname": "web01",
                  "consumer_id": 0,
                  "backups_enabled": false,
                  "password": "a8hjsjnbs91",
                  "label": "foo",
                  "slices_count": null,
                  "created_at": "2009-04-16T08:17:39-04:00"
                }
              }"""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _virtual_machines_api10json_create(self, method, url, body, headers):
        body = """{
              "virtual_machine":
                {
                  "running": true,
                  "updated_at": "2009-05-15T06:55:02-04:00",
                  "power_action_pending": false,
                  "system_template_id": 41,
                  "id": 1384,
                  "cloud_id": 3,
                  "domain_name": "demodomain.com",
                  "hostname": "web01",
                  "consumer_id": 0,
                  "backups_enabled": false,
                  "password": "a8hjsjnbs91",
                  "label": "foo",
                  "slices_count": null,
                  "created_at": "2009-04-16T08:17:39-04:00"
                }
              }"""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _virtual_machines_api10json_virtual_machines(self, method, url, body, headers):
        body = """     [{
              "virtual_machine":
                {
                  "running": true,
                  "updated_at": "2009-05-15T06:55:02-04:00",
                  "power_action_pending": false,
                  "system_template_id": 41,
                  "id": 1384,
                  "cloud_id": 3,
                  "domain_name": "demodomain.com",
                  "hostname": "web01",
                  "consumer_id": 0,
                  "backups_enabled": false,
                  "password": "a8hjsjnbs91",
                  "label": "Web Server 01",
                  "slices_count": null,
                  "created_at": "2009-04-16T08:17:39-04:00"
                }
              },
              {
                "virtual_machine":
                  {
                    "running": true,
                    "updated_at": "2009-05-15T06:55:02-04:00",
                    "power_action_pending": false,
                    "system_template_id": 41,
                    "id": 1385,
                    "cloud_id": 3,
                    "domain_name": "demodomain.com",
                    "hostname": "mysql01",
                    "consumer_id": 0,
                    "backups_enabled": false,
                    "password": "dsi8h38hd2s",
                    "label": "MySQL Server 01",
                    "slices_count": null,
                    "created_at": "2009-04-16T08:17:39-04:00"
                  }
                }]"""
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _available_clouds_api10json_templates(self, method, url, body, headers):
        body = self.fixtures.load("_available_clouds_api10json_templates.json")
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _available_clouds_api10json_create(self, method, url, body, headers):
        body = """
        [{"cloud":{"system_templates":[{"id":9,"label":"Ubuntu 8.04 x64"}],"id":2,"label":"USA VPS Cloud"}}]
        """
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest

from libcloud.test import MockHttp
from libcloud.utils.py3 import httplib, urlparse, parse_qsl
from libcloud.test.compute import TestCaseMixin
from libcloud.test.file_fixtures import ComputeFileFixtures
from libcloud.compute.drivers.ktucloud import KTUCloudNodeDriver

try:
    import simplejson as json
except ImportError:
    import json


class KTUCloudNodeDriverTest(unittest.TestCase, TestCaseMixin):
    def setUp(self):
        KTUCloudNodeDriver.connectionCls.conn_class = KTUCloudStackMockHttp
        self.driver = KTUCloudNodeDriver(
            "apikey", "secret", path="/test/path", host="api.dummy.com"
        )
        self.driver.path = "/test/path"
        self.driver.type = -1
        KTUCloudStackMockHttp.fixture_tag = "default"
        self.driver.connection.poll_interval = 0.0

    def test_create_node_immediate_failure(self):
        size = self.driver.list_sizes()[0]
        image = self.driver.list_images()[0]
        KTUCloudStackMockHttp.fixture_tag = "deployfail"
        self.assertRaises(
            Exception, self.driver.create_node, name="node-name", image=image, size=size
        )

    def test_create_node_delayed_failure(self):
        size = self.driver.list_sizes()[0]
        image = self.driver.list_images()[0]
        KTUCloudStackMockHttp.fixture_tag = "deployfail2"
        self.assertRaises(
            Exception, self.driver.create_node, name="node-name", image=image, size=size
        )

    def test_list_images_no_images_available(self):
        KTUCloudStackMockHttp.fixture_tag = "notemplates"

        images = self.driver.list_images()
        self.assertEqual(0, len(images))

    def test_list_images_available(self):
        images = self.driver.list_images()
        self.assertEqual(112, len(images))

    def test_list_sizes_available(self):
        sizes = self.driver.list_sizes()
        self.assertEqual(112, len(sizes))

    def test_list_sizes_nodisk(self):
        KTUCloudStackMockHttp.fixture_tag = "nodisk"

        sizes = self.driver.list_sizes()
        self.assertEqual(2, len(sizes))

        check = False
        size = sizes[1]
        if size.id == KTUCloudNodeDriver.EMPTY_DISKOFFERINGID:
            check = True

        self.assertTrue(check)


class KTUCloudStackMockHttp(MockHttp, unittest.TestCase):
    fixtures = ComputeFileFixtures("ktucloud")
    fixture_tag = "default"

    def _load_fixture(self, fixture):
        body = self.fixtures.load(fixture)
        return body, json.loads(body)

    def _test_path(self, method, url, body, headers):
        url = urlparse.urlparse(url)
        query = dict(parse_qsl(url.query))

        self.assertTrue("apiKey" in query)
        self.assertTrue("command" in query)
        self.assertTrue("response" in query)
        self.assertTrue("signature" in query)

        self.assertTrue(query["response"] == "json")

        del query["apiKey"]
        del query["response"]
        del query["signature"]
        command = query.pop("command")

        if hasattr(self, "_cmd_" + command):
            return getattr(self, "_cmd_" + command)(**query)
        else:
            fixture = command + "_" + self.fixture_tag + ".json"
            body, obj = self._load_fixture(fixture)
            return (httplib.OK, body, {}, httplib.responses[httplib.OK])

    def _cmd_queryAsyncJobResult(self, jobid):
        fixture = "queryAsyncJobResult" + "_" + str(jobid) + ".json"
        body, obj = self._load_fixture(fixture)
        return (httplib.OK, body, {}, httplib.responses[httplib.OK])


if __name__ == "__main__":
    sys.exit(unittest.main())
